from __future__ import annotations

from paperqa import PromptCollection

# I wanted to to make this an Enum
# but there is a metaclass conflict
# so we must instead have some logic here
# and some logic on named_prompt in the QueryRequest model
# https://github.com/pydantic/pydantic/issues/2173
STATIC_PROMPTS: dict[str, PromptCollection] = {
    "default": PromptCollection(
        qa=(
            "Answer the question below with the context.\n\n"
            "Context:\n\n{context}\n\n----\n\n"
            "Question: {question}\n\n"
            "Write an answer based on the context. "
            "If the context provides insufficient information and "
            "the question cannot be directly answered, reply "
            '"I cannot answer." '
            "For each part of your answer, indicate which sources most support "
            "it via citation keys at the end of sentences, "
            "like (Example2012Example pages 3-4). Only cite from the context "
            "below and only use the valid keys. "
            "Write in the style of a direct email containing only key details, equations, and quantities. "
            'Avoid using adverb phrases like "furthermore", "additionally", and "moreover." '
            "This will go directly onto a website for public viewing, so do not include any "
            "process details about following these instructions.\n\n"
            "Answer ({answer_length}):\n"
        ),
        select=(
            "Select papers that may help answer the question below. "
            "Papers are listed as $KEY: $PAPER_INFO. "
            "Return a list of keys, separated by commas. "
            'Return "None", if no papers are applicable. '
            "Choose papers that are relevant, from reputable sources, and timely "
            "(if the question requires timely information). \n\n"
            "Question: {question}\n\n"
            "Papers: {papers}\n\n"
            "Selected keys:"
        ),
        pre=(
            "We are collecting background information for the question/task below. "
            "Provide a brief summary of definitions, acronyms, or background information (about 50 words) that "
            "could help answer the question. Do not answer it directly. Ignore formatting instructions. "
            "Do not answer if there is nothing to contribute. "
            "\n\nQuestion:\n{question}\n\n"
        ),
        post=None,
        system=(
            "Answer in a direct and concise tone. "
            "Your audience is an expert, so be highly specific. "
            "If there are ambiguous terms or acronyms, be explicit."
        ),
        skip_summary=False,
        json_summary=True,
        summary_json=(
            "Excerpt from {citation}\n\n----\n\n{text}\n\n----\n\n"
            "Query: {question}\n\n"
        ),
        summary_json_system="Provide a summary of the excerpt that could help answer the question based on the excerpt.  "  # noqa: E501
        "The excerpt may be irrelevant. Do not directly answer the question - only summarize relevant information. "
        "Respond with the following JSON format:\n\n"
        '{{\n"summary": "...",\n"relevance_score": "..."}}\n\n'
        "where `summary` is relevant information from text ({summary_length}), "
        "and `relevance_score` is "
        "the relevance of `summary` to answer the question (integer out of 10).",
    ),
    "wikicrow": PromptCollection(
        qa=(
            "Answer the question below with the context.\n\n"
            "Context:\n\n{context}\n\n----\n\n"
            "Question: {question}\n\n"
            "Write an answer based on the context. "
            "If the context provides insufficient information and "
            "the question cannot be directly answered, reply "
            '"I cannot answer." '
            "For each part of your answer, indicate which sources most support "
            "it via citation keys at the end of sentences, "
            "like (Example2012Example pages 3-4). Only cite from the context "
            "below and only use the valid keys. Write in the style of a "
            "Wikipedia article, with concise sentences and coherent paragraphs. "
            "The context comes from a variety of sources and is only a summary, "
            "so there may inaccuracies or ambiguities. Make sure the gene_names exactly match "
            "the gene name in the question before using a context. "
            "This answer will go directly onto "
            "Wikipedia, so do not add any extraneous information.\n\n"
            "Answer ({answer_length}):"
        ),
        select=(
            "Select papers that may help answer the question below. "
            "Papers are listed as $KEY: $PAPER_INFO. "
            "Return a list of keys, separated by commas. "
            'Return "None", if no papers are applicable. '
            "Choose papers that are relevant, from reputable sources, and timely "
            "(if the question requires timely information). \n\n"
            "Question: {question}\n\n"
            "Papers: {papers}\n\n"
            "Selected keys:"
        ),
        pre=(
            "We are collecting background information for the question/task below. "
            "Provide a brief summary of definitions, acronyms, or background information (about 50 words) that "
            "could help answer the question. Do not answer it directly. Ignore formatting instructions. "
            "Do not answer if there is nothing to contribute. "
            "\n\nQuestion:\n{question}\n\n"
        ),
        post=None,
        system=(
            "Answer in a direct and concise tone. "
            "Your audience is an expert, so be highly specific. "
            "If there are ambiguous terms or acronyms, be explicit."
        ),
        skip_summary=False,
        json_summary=True,
        summary_json=(
            "Excerpt from {citation}\n\n----\n\n{text}\n\n----\n\n"
            "Query: {question}\n\n"
        ),
        summary_json_system="Provide a summary of the excerpt that could help answer the question based on the excerpt.  "  # noqa: E501
        "The excerpt may be irrelevant. Do not directly answer the question - only summarize relevant information. "
        "Respond with the following JSON format:\n\n"
        '{{\n"summary": "...",\n"gene_name: "...",\n"relevance_score": "..."}}\n\n'
        "where `summary` is relevant information from text ({summary_length}), "
        "`gene_name` is the gene discussed in the excerpt (may be different than query), "
        "and `relevance_score` is "
        "the relevance of `summary` to answer the question (integer out of 10).",
    ),
}

# for backwards compatibility
STATIC_PROMPTS["json"] = STATIC_PROMPTS["default"]



from __future__ import annotations

import ast
import logging
import operator
import os
import shutil
from datetime import datetime
from pathlib import Path
from typing import Any

import yaml
from typing_extensions import Annotated

from .. import __version__
from ..utils import get_loop, pqa_directory

try:
    import anyio
    import typer
    from rich.console import Console
    from rich.logging import RichHandler

    from .main import agent_query, search
    from .models import AnswerResponse, MismatchedModelsError, QueryRequest
    from .search import SearchIndex, get_directory_index

except ImportError as e:
    raise ImportError(
        '"agents" module is not installed please install it using "pip install paper-qa[agents]"'
    ) from e

logger = logging.getLogger(__name__)

app = typer.Typer()


def configure_agent_logging(
    verbosity: int = 0, default_level: int = logging.INFO
) -> None:
    """Default to INFO level, but suppress loquacious loggers."""
    verbosity_map = {
        0: {
            "paperqa.agents.helpers": logging.WARNING,
            "paperqa.agents.main": logging.WARNING,
            "anthropic": logging.WARNING,
            "openai": logging.WARNING,
            "httpx": logging.WARNING,
            "paperqa.agents.models": logging.WARNING,
        }
    }

    verbosity_map[1] = verbosity_map[0] | {
        "paperqa.agents.main": logging.INFO,
        "paperqa.models": logging.INFO,
    }

    verbosity_map[2] = verbosity_map[1] | {
        "paperqa.agents.helpers": logging.DEBUG,
        "paperqa.agents.main": logging.DEBUG,
        "paperqa.agents.main.agent_callers": logging.DEBUG,
        "paperqa.models": logging.DEBUG,
        "paperqa.agents.search": logging.DEBUG,
    }

    rich_handler = RichHandler(
        rich_tracebacks=True,
        markup=True,
        show_path=False,
        show_level=False,
        console=Console(force_terminal=True),
    )

    rich_handler.setFormatter(logging.Formatter("%(message)s", datefmt="[%X]"))

    def is_paperqa_related(logger_name: str) -> bool:
        return logger_name.startswith("paperqa") or logger_name in {
            "anthropic",
            "openai",
            "httpx",
        }

    for logger_name, logger in logging.Logger.manager.loggerDict.items():
        if isinstance(logger, logging.Logger) and is_paperqa_related(logger_name):
            logger.setLevel(
                verbosity_map.get(min(verbosity, 2), {}).get(logger_name, default_level)
            )
            if not any(isinstance(h, RichHandler) for h in logger.handlers):
                logger.addHandler(rich_handler)


def get_file_timestamps(path: os.PathLike | str) -> dict[str, str]:
    # Get the stats for the file/directory
    stats = os.stat(path)

    # Get created time (ctime)
    created_time = datetime.fromtimestamp(stats.st_ctime)

    # Get modified time (mtime)
    modified_time = datetime.fromtimestamp(stats.st_mtime)

    return {
        "created_at": created_time.strftime("%Y-%m-%d %H:%M:%S"),
        "modified_at": modified_time.strftime("%Y-%m-%d %H:%M:%S"),
    }


def parse_dot_to_dict(str_w_dots: str, value: str) -> dict[str, Any]:
    parsed: dict[str, Any] = {}
    for key in str_w_dots.split(".")[::-1]:
        if not parsed:
            try:
                eval_value = ast.literal_eval(value)
                if isinstance(eval_value, (set, list)):
                    parsed[key] = eval_value
                else:
                    parsed[key] = value
            except (ValueError, SyntaxError):
                parsed[key] = value
        else:
            parsed = {key: parsed}
    return parsed


def pop_nested_dict_recursive(d: dict[str, Any], path: str) -> tuple[Any, bool]:
    """
    Pop a value from a nested dictionary (in-place) using a period-separated path.

    Recursively remove empty dictionaries after popping.
    """
    keys = path.split(".")

    if len(keys) == 1:
        if keys[0] not in d:
            raise KeyError(f"Key not found: {keys[0]}")
        value = d.pop(keys[0])
        return value, len(d) == 0

    if keys[0] not in d or not isinstance(d[keys[0]], dict):
        raise KeyError(f"Invalid path: {path}")

    value, should_remove = pop_nested_dict_recursive(d[keys[0]], ".".join(keys[1:]))

    if should_remove:
        d.pop(keys[0])

    return value, len(d) == 0


def get_settings(
    settings_path: str | os.PathLike | None = None,
) -> dict[str, Any]:

    if settings_path is None:
        settings_path = pqa_directory("settings") / "settings.yaml"

    if os.path.exists(settings_path):
        with open(settings_path) as f:
            return yaml.safe_load(f)

    return {}


def merge_dicts(dict_a: dict, dict_b: dict) -> dict:
    """
    Merge two dictionaries where if dict_a has a key with a subdictionary.

    dict_b only overwrites the keys in dict_a's subdictionary if they are
    also specified in dict_b, but otherwise keeps all the subkeys.
    """
    result = dict_a.copy()  # Start with a shallow copy of dict_a

    for key, value in dict_b.items():
        if isinstance(value, dict) and key in result and isinstance(result[key], dict):
            # If both dict_a and dict_b have a dict for this key, recurse
            result[key] = merge_dicts(result[key], value)
        else:
            # Otherwise, just update the value
            result[key] = value

    return result


def get_merged_settings(
    settings: dict[str, Any], settings_path: Path | None = None
) -> dict[str, Any]:
    """Merges a new settings with the current settings saved to file."""
    current_settings = get_settings(settings_path)

    # deal with the nested key case
    return merge_dicts(current_settings, settings)


@app.command("set")
def set_setting(
    variable: Annotated[
        str,
        typer.Argument(
            help=(
                "PaperQA variable to set, see agents.models.QueryRequest object for all settings, "
                "nested options can be set using periods, ex. agent_tools.paper_directory"
            )
        ),
    ],
    value: Annotated[
        str,
        typer.Argument(
            help=(
                "Value to set to the variable, will be cast to the correct type automatically."
            )
        ),
    ],
) -> None:
    """Set a persistent PaperQA setting."""
    configure_agent_logging(verbosity=0)

    settings_path = pqa_directory("settings") / "settings.yaml"

    current_settings = get_merged_settings(
        parse_dot_to_dict(variable, value), settings_path=settings_path
    )

    try:
        QueryRequest(**current_settings)
    except MismatchedModelsError:
        pass
    except ValueError as e:
        raise ValueError(
            f"{variable} (with value {value}) is not a valid setting."
        ) from e

    logger.info(f"{variable} set to {str(value)[:100]}!")

    with open(settings_path, "w") as f:
        yaml.dump(current_settings, f)


@app.command()
def show(
    variable: Annotated[
        str,
        typer.Argument(
            help=(
                "PaperQA variable to show, see agents.models.QueryRequest object for all settings, "
                "nested options can be set using periods, ex. agent_tools.paper_directory. "
                "Can show all indexes with `indexes` input, answers with `answers` input, "
                "and `all` for all settings."
            )
        ),
    ],
    limit: Annotated[
        int, typer.Option(help="limit results, only used for 'answers'.")
    ] = 5,
) -> Any:
    """Show a persistent PaperQA setting, special inputs include `indexes`, `answers` and `all`."""
    configure_agent_logging(verbosity=0)

    # handle special case when user wants to see indexes
    if variable == "indexes":
        for index in os.listdir(pqa_directory("indexes")):
            index_times = get_file_timestamps(pqa_directory("indexes") / index)
            logger.info(f"{index}, {index_times}")
        return os.listdir(pqa_directory("indexes"))

    if variable == "answers":
        all_answers = []
        answer_file_location = pqa_directory("indexes") / "answers" / "docs"
        if os.path.exists(answer_file_location):
            for answer_file in os.listdir(answer_file_location):
                all_answers.append(
                    get_file_timestamps(os.path.join(answer_file_location, answer_file))
                )
                with open(os.path.join(answer_file_location, answer_file)) as f:
                    answer = yaml.safe_load(f)
                all_answers[-1].update({"answer": answer})
            all_answers = sorted(
                all_answers, key=operator.itemgetter("modified_at"), reverse=True
            )[:limit]
            for answer in all_answers:
                logger.info(
                    f"Q: {answer['answer']['answer']['question']}\n---\nA: {answer['answer']['answer']['answer']}\n\n\n"
                )
        return all_answers

    current_settings = get_settings(pqa_directory("settings") / "settings.yaml")

    if variable == "all":
        logger.info(current_settings)
        return current_settings

    try:
        value, _ = pop_nested_dict_recursive(current_settings, variable)
    except KeyError:
        logger.info(f"{variable} is not set.")
        return None
    else:
        logger.info(f"{variable}: {value}")
        return value


@app.command()
def clear(
    variable: Annotated[
        str,
        typer.Argument(
            help=(
                "PaperQA variable to clear, see agents.models.QueryRequest object for all settings, "
                "nested options can be set using periods, ex. agent_tools.paper_directory. "
                "Index names can also be used if the --index flag is set."
            )
        ),
    ],
    index: Annotated[
        bool,
        typer.Option(
            "--index",
            is_flag=True,
            help="index flag to indicate that this index name should be cleared.",
        ),
    ] = False,
) -> None:
    """Clear a persistent PaperQA setting, include the --index flag to remove an index."""
    configure_agent_logging(verbosity=0)

    settings_path = pqa_directory("settings") / "settings.yaml"

    current_settings = get_settings(settings_path)

    if not index:
        _ = pop_nested_dict_recursive(current_settings, variable)
        with open(settings_path, "w") as f:
            yaml.dump(current_settings, f)
        logger.info(f"{variable} cleared!")

    elif variable in os.listdir(pqa_directory("indexes")):
        shutil.rmtree(pqa_directory("indexes") / variable)
        logger.info(f"Index {variable} cleared!")
    else:
        logger.info(f"Index {variable} not found!")


@app.command()
def ask(
    query: Annotated[str, typer.Argument(help=("Question or task ask of PaperQA"))],
    agent_type: Annotated[
        str,
        typer.Option(
            help=(
                "Type of agent to use, for now either "
                "`OpenAIFunctionsAgent` or `fake`. `fake` uses "
                "a hard coded tool path (search->gather evidence->answer)."
            )
        ),
    ] = "fake",
    verbosity: Annotated[
        int, typer.Option(help=("Level of verbosity from 0->2 (inclusive)"))
    ] = 0,
    directory: Annotated[
        Path | None,
        typer.Option(help=("Directory of papers or documents to run PaperQA over.")),
    ] = None,
    index_directory: Annotated[
        Path | None,
        typer.Option(
            help=(
                "Index directory to store paper index and answers. Default will be `~/.pqa`"
            )
        ),
    ] = None,
    manifest_file: Annotated[
        Path | None,
        typer.Option(
            help=(
                "Optional manifest file (CSV) location to map relative a "
                "`file_location` column to `doi` or `title` columns. "
                "If not used, then the file will be read by an LLM "
                "which attempts to extract the title, authors and DOI."
            )
        ),
    ] = None,
) -> AnswerResponse:
    """Query PaperQA via an agent."""
    configure_agent_logging(verbosity=verbosity)

    loop = get_loop()

    # override settings file if requested directly
    to_merge = {}

    if directory is not None:
        to_merge = {"agent_tools": {"paper_directory": directory}}

    if index_directory is not None:
        if "agent_tools" not in to_merge:
            to_merge = {"agent_tools": {"index_directory": index_directory}}
        else:
            to_merge["agent_tools"].update({"index_directory": index_directory})

    if manifest_file is not None:
        if "agent_tools" not in to_merge:
            to_merge = {"agent_tools": {"manifest_file": manifest_file}}
        else:
            to_merge["agent_tools"].update({"manifest_file": manifest_file})

    request = QueryRequest(
        query=query,
        **get_merged_settings(
            to_merge,
            settings_path=pqa_directory("settings") / "settings.yaml",
        ),
    )

    return loop.run_until_complete(
        agent_query(
            request,
            docs=None,
            verbosity=verbosity,
            agent_type=agent_type,
            index_directory=request.agent_tools.index_directory,
        )
    )


@app.command("search")
def search_query(
    query: Annotated[str, typer.Argument(help=("Query for keyword search"))],
    index_name: Annotated[
        str,
        typer.Argument(
            help=(
                "Name of the index to search, or use `answers`"
                " to search all indexed answers"
            )
        ),
    ] = "answers",
    index_directory: Annotated[
        Path | None,
        typer.Option(
            help=(
                "Index directory to store paper index and answers. Default will be `~/.pqa`"
            )
        ),
    ] = None,
) -> list[tuple[AnswerResponse, str] | tuple[Any, str]]:
    """Search using a pre-built PaperQA index."""
    configure_agent_logging(verbosity=0)

    loop = get_loop()
    return loop.run_until_complete(
        search(
            query,
            index_name=index_name,
            index_directory=index_directory or pqa_directory("indexes"),
        )
    )


@app.command("index")
def build_index(
    directory: Annotated[
        Path | None,
        typer.Argument(help=("Directory of papers or documents to run PaperQA over.")),
    ] = None,
    index_directory: Annotated[
        Path | None,
        typer.Option(
            help=(
                "Index directory to store paper index and answers. Default will be `~/.pqa`"
            )
        ),
    ] = None,
    manifest_file: Annotated[
        Path | None,
        typer.Option(
            help=(
                "Optional manifest file (CSV) location to map relative a "
                "`file_location` column to `doi` or `title` columns. "
                "If not used, then the file will be read by an LLM "
                "which attempts to extract the title, authors and DOI."
            )
        ),
    ] = None,
    verbosity: Annotated[
        int, typer.Option(help=("Level of verbosity from 0->2 (inclusive)"))
    ] = 0,
) -> SearchIndex:
    """Build a PaperQA search index, this will also happen automatically upon using `ask`."""
    configure_agent_logging(verbosity=verbosity)

    to_merge = {}

    if directory is not None:
        to_merge = {"agent_tools": {"paper_directory": directory}}

    if index_directory is not None:
        if "agent_tools" not in to_merge:
            to_merge = {"agent_tools": {"index_directory": index_directory}}
        else:
            to_merge["agent_tools"].update({"index_directory": index_directory})

    if manifest_file is not None:
        if "agent_tools" not in to_merge:
            to_merge = {"agent_tools": {"manifest_file": manifest_file}}
        else:
            to_merge["agent_tools"].update({"manifest_file": manifest_file})

    configure_agent_logging(verbosity)

    request_settings = QueryRequest(
        query="",
        **get_merged_settings(
            to_merge,
            settings_path=pqa_directory("settings") / "settings.yaml",
        ),
    )

    loop = get_loop()

    return loop.run_until_complete(
        get_directory_index(
            directory=anyio.Path(request_settings.agent_tools.paper_directory),
            index_directory=request_settings.agent_tools.index_directory,
            index_name=request_settings.get_index_name(
                request_settings.agent_tools.paper_directory,
                request_settings.embedding,
                request_settings.parsing_configuration,
            ),
            manifest_file=(
                anyio.Path(request_settings.agent_tools.manifest_file)
                if request_settings.agent_tools.manifest_file
                else None
            ),
            embedding=request_settings.embedding,
            chunk_chars=request_settings.parsing_configuration.chunksize,
            overlap=request_settings.parsing_configuration.overlap,
        )
    )


@app.command()
def version():
    configure_agent_logging(verbosity=0)
    logger.info(f"PaperQA version: {__version__}")


if __name__ == "__main__":
    app()



from __future__ import annotations

import logging
import os
import re
from datetime import datetime
from typing import Any, Callable, ClassVar, Collection
from uuid import UUID, uuid4

import tiktoken
from pybtex.database import BibliographyData, Entry, Person
from pybtex.database.input.bibtex import Parser
from pybtex.scanner import PybtexSyntaxError
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    computed_field,
    field_validator,
    model_validator,
)

from .prompts import (
    citation_prompt,
    default_system_prompt,
    qa_prompt,
    select_paper_prompt,
    structured_citation_prompt,
    summary_json_prompt,
    summary_json_system_prompt,
    summary_prompt,
)
from .utils import (
    create_bibtex_key,
    encode_id,
    format_bibtex,
    get_citenames,
)
from .version import __version__ as pqa_version

# Just for clarity
DocKey = Any
CallbackFactory = Callable[[str], list[Callable[[str], None]] | None]

logger = logging.getLogger(__name__)


class LLMResult(BaseModel):
    """A class to hold the result of a LLM completion."""

    id: UUID = Field(default_factory=uuid4)
    answer_id: UUID | None = None
    name: str | None = None
    prompt: str | list[dict] | None = Field(
        default=None,
        description="Optional prompt (str) or list of serialized prompts (list[dict]).",
    )
    text: str = ""
    prompt_count: int = 0
    completion_count: int = 0
    model: str
    date: str = Field(default_factory=datetime.now().isoformat)
    seconds_to_first_token: float = Field(
        default=0.0, description="Delta time (sec) to first response token's arrival."
    )
    seconds_to_last_token: float = Field(
        default=0.0, description="Delta time (sec) to last response token's arrival."
    )

    def __str__(self):
        return self.text


class Embeddable(BaseModel):
    embedding: list[float] | None = Field(default=None, repr=False)


class Doc(Embeddable):
    docname: str
    citation: str
    dockey: DocKey
    overwrite_fields_from_metadata: bool = Field(
        default=True,
        description="flag to overwrite fields from metadata when upgrading to a DocDetails",
    )

    def __hash__(self) -> int:
        return hash((self.docname, self.dockey))


class Text(Embeddable):
    text: str
    name: str
    doc: Doc


# Mock a dictionary and store any missing items
class _FormatDict(dict):
    def __init__(self) -> None:
        self.key_set: set[str] = set()

    def __missing__(self, key: str) -> str:
        self.key_set.add(key)
        return key


def get_formatted_variables(s: str) -> set[str]:
    """Returns the set of variables implied by the format string."""
    format_dict = _FormatDict()
    s.format_map(format_dict)
    return format_dict.key_set


class PromptCollection(BaseModel):
    summary: str = summary_prompt
    qa: str = qa_prompt
    select: str = select_paper_prompt
    cite: str = citation_prompt
    structured_cite: str = structured_citation_prompt
    pre: str | None = Field(
        default=None,
        description=(
            "Opt-in pre-prompt (templated with just the original question) to append"
            " information before a qa prompt. For example:"
            " 'Summarize all scientific terms in the following question:\n{question}'."
            " This pre-prompt can enable injection of question-specific guidance later"
            " used by the qa prompt, without changing the qa prompt's template."
        ),
    )
    post: str | None = None
    system: str = default_system_prompt
    skip_summary: bool = False
    json_summary: bool = False
    # Not thrilled about this model,
    # but need to split out the system/summary
    # to get JSON
    summary_json: str = summary_json_prompt
    summary_json_system: str = summary_json_system_prompt

    @field_validator("summary")
    @classmethod
    def check_summary(cls, v: str) -> str:
        if not set(get_formatted_variables(v)).issubset(
            set(get_formatted_variables(summary_prompt))
        ):
            raise ValueError(
                f"Summary prompt can only have variables: {get_formatted_variables(summary_prompt)}"
            )
        return v

    @field_validator("qa")
    @classmethod
    def check_qa(cls, v: str) -> str:
        if not set(get_formatted_variables(v)).issubset(
            set(get_formatted_variables(qa_prompt))
        ):
            raise ValueError(
                f"QA prompt can only have variables: {get_formatted_variables(qa_prompt)}"
            )
        return v

    @field_validator("select")
    @classmethod
    def check_select(cls, v: str) -> str:
        if not set(get_formatted_variables(v)).issubset(
            set(get_formatted_variables(select_paper_prompt))
        ):
            raise ValueError(
                f"Select prompt can only have variables: {get_formatted_variables(select_paper_prompt)}"
            )
        return v

    @field_validator("pre")
    @classmethod
    def check_pre(cls, v: str | None) -> str | None:
        if v is not None and set(get_formatted_variables(v)) != {"question"}:
            raise ValueError("Pre prompt must have input variables: question")
        return v

    @field_validator("post")
    @classmethod
    def check_post(cls, v: str | None) -> str | None:
        if v is not None:
            # kind of a hack to get list of attributes in answer
            attrs = set(Answer.model_fields.keys())
            if not set(get_formatted_variables(v)).issubset(attrs):
                raise ValueError(f"Post prompt must have input variables: {attrs}")
        return v


class Context(BaseModel):
    """A class to hold the context of a question."""

    model_config = ConfigDict(extra="allow")

    context: str = Field(description="Summary of the text with respect to a question.")
    text: Text
    score: int = 5


def __str__(self) -> str:  # noqa: N807
    """Return the context as a string."""
    return self.context


class Answer(BaseModel):
    """A class to hold the answer to a question."""

    id: UUID = Field(default_factory=uuid4)
    question: str
    answer: str = ""
    context: str = ""
    contexts: list[Context] = []
    references: str = ""
    formatted_answer: str = ""
    dockey_filter: set[DocKey] | None = None
    summary_length: str = "about 100 words"
    answer_length: str = "about 100 words"
    # just for convenience you can override this
    cost: float | None = None
    # Map model name to a two-item list of LLM prompt token counts
    # and LLM completion token counts
    token_counts: dict[str, list[int]] = Field(default_factory=dict)
    model_config = ConfigDict(extra="ignore")

    def __str__(self) -> str:
        """Return the answer as a string."""
        return self.formatted_answer

    @model_validator(mode="before")
    @classmethod
    def remove_computed(cls, data: Any) -> Any:
        if isinstance(data, dict):
            data.pop("used_contexts", None)
        return data

    @computed_field  # type: ignore[misc]
    @property
    def used_contexts(self) -> set[str]:
        """Return the used contexts."""
        return get_citenames(self.formatted_answer)

    def get_citation(self, name: str) -> str:
        """Return the formatted citation for the given docname."""
        try:
            doc: Doc = next(
                filter(lambda x: x.text.name == name, self.contexts)
            ).text.doc
        except StopIteration as exc:
            raise ValueError(f"Could not find docname {name} in contexts.") from exc
        return doc.citation

    def add_tokens(self, result: LLMResult):
        """Update the token counts for the given result."""
        if result.model not in self.token_counts:
            self.token_counts[result.model] = [
                result.prompt_count,
                result.completion_count,
            ]
        else:
            self.token_counts[result.model][0] += result.prompt_count
            self.token_counts[result.model][1] += result.completion_count

    def get_unique_docs_from_contexts(self, score_threshold: int = 0) -> set[Doc]:
        """Parse contexts for docs with scores above the input threshold."""
        return {
            c.text.doc
            for c in filter(lambda x: x.score >= score_threshold, self.contexts)
        }

    def filter_content_for_user(self) -> None:
        """Filter out extra items (inplace) that do not need to be returned to the user."""
        self.contexts = [
            Context(
                context=c.context,
                score=c.score,
                text=Text(
                    text="",
                    **c.text.model_dump(exclude={"text", "embedding", "doc"}),
                    doc=Doc(**c.text.doc.model_dump(exclude={"embedding"})),
                ),
            )
            for c in self.contexts
        ]


class ChunkMetadata(BaseModel):
    """Metadata for chunking algorithm."""

    chunk_chars: int
    overlap: int
    chunk_type: str


class ParsedMetadata(BaseModel):
    """Metadata for parsed text."""

    parsing_libraries: list[str]
    total_parsed_text_length: int
    paperqa_version: str = pqa_version
    parse_type: str | None = None
    chunk_metadata: ChunkMetadata | None = None


class ParsedText(BaseModel):
    """Parsed text (pre-chunking)."""

    content: dict | str | list[str]
    metadata: ParsedMetadata

    def encode_content(self):
        # we tokenize using tiktoken so cuts are in reasonable places
        # See https://github.com/openai/tiktoken
        enc = tiktoken.get_encoding("cl100k_base")
        if isinstance(self.content, str):
            return enc.encode_ordinary(self.content)
        elif isinstance(self.content, list):  # noqa: RET505
            return [enc.encode_ordinary(c) for c in self.content]
        else:
            raise NotImplementedError(
                "Encoding only implemented for str and list[str] content."
            )


# We use these integer values
# as defined in https://jfp.csc.fi/en/web/haku/kayttoohje
# which is a recommended ranking system
SOURCE_QUALITY_MESSAGES = {
    0: "poor quality or predatory journal",
    1: "peer-reviewed journal",
    2: "domain leading peer-reviewed journal",
    3: "highest quality peer-reviewed journal",
}

CITATION_FALLBACK_DATA = {
    "authors": ["Unknown authors"],
    "author": "Unknown author(s)",
    "year": "Unknown year",
    "title": "Unknown title",
    "journal": "Unknown journal",
}


class DocDetails(Doc):
    model_config = ConfigDict(validate_assignment=True)

    citation: str
    key: str | None = None
    bibtex: str | None = None
    authors: list[str] | None = None
    publication_date: datetime | None = None
    year: int | None = None
    volume: str | None = None
    issue: str | None = None  # TODO: in bibtex this may be "number"
    issn: str | None = None
    pages: str | None = None
    journal: str | None = None
    publisher: str | None = None
    url: str | None = Field(
        default=None,
        description=(
            "Optional URL to the paper, which can lead to a Semantic Scholar page,"
            " arXiv abstract, etc. As of version 0.67 on 5/10/2024, we don't use this"
            " URL anywhere in the source code."
        ),
    )
    title: str | None = None
    citation_count: int | None = None
    bibtex_type: str | None = None

    source_quality: int | None = Field(
        default=None,
        description="Quality of journal/venue of paper. "
        " We use None as a sentinel for unset values (like for determining hydration) "
        " So, we use -1 means unknown quality and None means it needs to be hydrated.",
    )
    doi: str | None = None
    doi_url: str | None = None
    doc_id: str | None = None
    file_location: str | os.PathLike | None = None
    other: dict[str, Any] = Field(
        default_factory=dict,
        description="Other metadata besides the above standardized fields.",
    )
    UNDEFINED_JOURNAL_QUALITY: ClassVar[int] = -1
    DOI_URL_FORMATS: ClassVar[Collection[str]] = {
        "https://doi.org/",
        "http://dx.doi.org/",
    }
    AUTHOR_NAMES_TO_REMOVE: ClassVar[Collection[str]] = {"et al", "et al."}

    @field_validator("key")
    @classmethod
    def clean_key(cls, value: str) -> str:
        # Replace HTML tags with empty string
        return re.sub(pattern=r"<\/?\w{1,10}>", repl="", string=value)

    @classmethod
    def lowercase_doi_and_populate_doc_id(cls, data: dict[str, Any]) -> dict[str, Any]:
        if doi := data.get("doi"):
            remove_urls = cls.DOI_URL_FORMATS
            for url in remove_urls:
                if doi.startswith(url):
                    doi = doi.replace(url, "")
            data["doi"] = doi.lower()
            data["doc_id"] = encode_id(doi.lower())
        else:
            data["doc_id"] = encode_id(uuid4())

        if data.get("overwrite_fields_from_metadata", True):
            data["dockey"] = data["doc_id"]

        return data

    @staticmethod
    def is_bibtex_complete(bibtex: str, fields: list[str] | None = None) -> bool:
        """Validate bibtex entries have certain fields."""
        if fields is None:
            fields = ["doi", "title"]
        return all(field + "=" in bibtex for field in fields)

    @staticmethod
    def merge_bibtex_entries(entry1: Entry, entry2: Entry) -> Entry:
        """Merge two bibtex entries into one, preferring entry2 fields."""
        merged_entry = Entry(entry1.type)

        for field, value in entry1.fields.items():
            merged_entry.fields[field] = value
        for field, value in entry2.fields.items():
            merged_entry.fields[field] = value

        return merged_entry

    @staticmethod
    def misc_string_cleaning(data: dict[str, Any]) -> dict[str, Any]:
        """Clean strings before the enter the validation process."""
        if pages := data.get("pages"):
            data["pages"] = pages.replace("--", "-").replace(" ", "")
        return data

    @staticmethod
    def inject_clean_doi_url_into_data(data: dict[str, Any]) -> dict[str, Any]:
        """Ensure doi_url is present in data (since non-default arguments are not included)."""
        doi_url, doi = data.get("doi_url"), data.get("doi")

        if doi and not doi_url:
            doi_url = "https://doi.org/" + doi

        # ensure the modern doi url is used
        if doi_url:
            data["doi_url"] = doi_url.replace(
                "http://dx.doi.org/", "https://doi.org/"
            ).lower()

        return data

    @classmethod
    def remove_invalid_authors(cls, data: dict[str, Any]) -> dict[str, Any]:
        """Capture and cull strange author names."""
        if authors := data.get("authors"):
            data["authors"] = [
                a for a in authors if a.lower() not in cls.AUTHOR_NAMES_TO_REMOVE
            ]

        return data

    @staticmethod
    def overwrite_docname_dockey_for_compatibility_w_doc(
        data: dict[str, Any]
    ) -> dict[str, Any]:
        """Overwrite fields from metadata if specified."""
        overwrite_fields = {"key": "docname", "doc_id": "dockey"}
        if data.get("overwrite_fields_from_metadata", True):
            for field, old_field in overwrite_fields.items():
                if data.get(field):
                    data[old_field] = data[field]
        return data

    @classmethod
    def populate_bibtex_key_citation(  # noqa: C901, PLR0912
        cls, data: dict[str, Any]
    ) -> dict[str, Any]:
        """Add or modify bibtex, key, and citation fields.

        Missing values, 'unknown' keys, and incomplete bibtex entries are regenerated.

        When overwrite_fields_from_metadata:
            If bibtex is regenerated, the citation field is also regenerated.

            Otherwise we keep the citation field as is.

        """
        # we try to regenerate the key if unknowns are present, maybe they have been found
        if not data.get("key") or "unknown" in data["key"].lower():
            data["key"] = create_bibtex_key(
                data.get("authors") or CITATION_FALLBACK_DATA["authors"],  # type: ignore[arg-type]
                data.get("year") or CITATION_FALLBACK_DATA["year"],  # type: ignore[arg-type]
                data.get("title") or CITATION_FALLBACK_DATA["title"],  # type: ignore[arg-type]
            )
            if data.get("overwrite_fields_from_metadata", True):
                data["docname"] = data["key"]

        # even if we have a bibtex, it may not be complete, thus we need to add to it
        if not data.get("bibtex") or not cls.is_bibtex_complete(data["bibtex"]):
            existing_entry = None
            # if our bibtex already exists, but is incomplete, we add self_generated to metadata
            if data.get("bibtex"):
                if data.get("other"):
                    if (
                        "bibtex_source" in data["other"]
                        and "self_generated" not in data["other"]["bibtex_source"]
                    ):
                        data["other"]["bibtex_source"].append("self_generated")
                    else:
                        data["other"]["bibtex_source"] = ["self_generated"]
                else:
                    data["other"] = {"bibtex_source": ["self_generated"]}
                try:
                    existing_entry = next(
                        iter(Parser().parse_string(data["bibtex"]).entries.values())
                    )
                except PybtexSyntaxError:
                    logger.warning(f"Failed to parse bibtex for {data['bibtex']}.")
                    existing_entry = None

            entry_data = {
                "title": data.get("title") or CITATION_FALLBACK_DATA["title"],
                "year": (
                    CITATION_FALLBACK_DATA["year"]
                    if not data.get("year")
                    else str(data["year"])
                ),
                "journal": data.get("journal") or CITATION_FALLBACK_DATA["journal"],
                "volume": data.get("volume"),
                "pages": data.get("pages"),
                "month": (
                    None
                    if not data.get("publication_date")
                    else data["publication_date"].strftime("%b")
                ),
                "doi": data.get("doi"),
                "url": data.get("doi_url"),
                "publisher": data.get("publisher"),
                "issue": data.get("issue"),
                "issn": data.get("issn"),
            }
            entry_data = {k: v for k, v in entry_data.items() if v}
            try:
                new_entry = Entry(
                    data.get("bibtex_type", "article") or "article", fields=entry_data
                )
                if existing_entry:
                    new_entry = cls.merge_bibtex_entries(existing_entry, new_entry)
                # add in authors manually into the entry
                authors = [Person(a) for a in data.get("authors", ["Unknown authors"])]
                for a in authors:
                    new_entry.add_person(a, "author")
                data["bibtex"] = BibliographyData(
                    entries={data["key"]: new_entry}
                ).to_string("bibtex")
                # clear out the citation, since it will be regenerated
                if data.get("overwrite_fields_from_metadata", True):
                    data["citation"] = None
            except Exception:
                logger.warning(
                    f"Failed to generate bibtex for {data.get('docname') or data.get('citation')}"
                )
        if not data.get("citation"):
            data["citation"] = format_bibtex(
                data["bibtex"], clean=True, missing_replacements=CITATION_FALLBACK_DATA  # type: ignore[arg-type]
            )
        return data

    @model_validator(mode="before")
    @classmethod
    def validate_all_fields(cls, data: dict[str, Any]) -> dict[str, Any]:
        data = cls.lowercase_doi_and_populate_doc_id(data)
        data = cls.remove_invalid_authors(data)
        data = cls.misc_string_cleaning(data)
        data = cls.inject_clean_doi_url_into_data(data)
        data = cls.populate_bibtex_key_citation(data)
        return cls.overwrite_docname_dockey_for_compatibility_w_doc(data)

    def __getitem__(self, item: str):
        """Allow for dictionary-like access, falling back on other."""
        try:
            return getattr(self, item)
        except AttributeError:
            return self.other[item]

    @property
    def formatted_citation(self) -> str:
        if (
            self.citation is None  # type: ignore[redundant-expr]
            or self.citation_count is None
            or self.source_quality is None
        ):
            raise ValueError(
                "Citation, citationCount, and sourceQuality are not set -- do you need to call `hydrate`?"
            )
        quality = (
            SOURCE_QUALITY_MESSAGES[self.source_quality]
            if self.source_quality >= 0
            else None
        )
        if quality is None:
            return f"{self.citation} This article has {self.citation_count} citations."
        return (
            f"{self.citation} This article has {self.citation_count} citations and is from a "
            f"{quality}."
        )

    OPTIONAL_HYDRATION_FIELDS: ClassVar[Collection[str]] = {"url"}

    def is_hydration_needed(  # pylint: disable=dangerous-default-value
        self,
        exclusion: Collection[str] = OPTIONAL_HYDRATION_FIELDS,
        inclusion: Collection[str] = [],
    ) -> bool:
        """Determine if we have unfilled attributes."""
        if inclusion:
            return any(
                v is None for k, v in self.model_dump().items() if k in inclusion
            )
        return any(
            v is None for k, v in self.model_dump().items() if k not in exclusion
        )

    def repopulate_doc_id_from_doi(self) -> None:
        # TODO: should this be a hash of the doi?
        if self.doi:
            self.doc_id = encode_id(self.doi)

    def __add__(self, other: DocDetails | int) -> DocDetails:  # noqa: C901
        """Merge two DocDetails objects together."""
        # control for usage w. Python's sum() function
        if isinstance(other, int):
            return self

        # first see if one of the entries is newer, which we will prefer
        PREFER_OTHER = True
        if self.publication_date and other.publication_date:
            PREFER_OTHER = self.publication_date <= other.publication_date

        merged_data = {}
        for field in self.model_fields:
            self_value = getattr(self, field)
            other_value = getattr(other, field)

            if field == "other":
                # Merge 'other' dictionaries
                merged_data[field] = {**self.other, **other.other}
                # handle the bibtex / sources as special fields
                for field_to_combine in ["bibtex_source", "client_source"]:
                    if self.other.get(field_to_combine) and other.other.get(
                        field_to_combine
                    ):
                        # Note: these should always be lists
                        merged_data[field][field_to_combine] = (
                            self.other[field_to_combine] + other.other[field_to_combine]
                        )

            elif field == "authors" and self_value and other_value:
                # Combine authors lists, removing duplicates
                # Choose whichever author list is longer
                best_authors = (
                    self.authors
                    if (
                        sum(len(a) for a in (self.authors or []))
                        >= sum(len(a) for a in (other.authors or []))
                    )
                    else other.authors
                )
                merged_data[field] = best_authors or None  # type: ignore[assignment]

            elif field == "key" and self_value is not None and other_value is not None:
                # if we have multiple keys, we wipe them and allow regeneration
                merged_data[field] = None  # type: ignore[assignment]

            elif (
                field in {"citation_count", "year", "publication_date"}
                and self_value is not None
                and other_value is not None
            ):
                # get the latest data
                merged_data[field] = max(self_value, other_value)

            else:
                # Prefer non-null values, default preference for 'other' object.
                # Note: if PREFER_OTHER = False then even if 'other' data exists
                # we will use 'self' data. This is to control for when we have
                # pre-prints / arXiv versions of papers that are not as up-to-date
                merged_data[field] = (
                    other_value
                    if (
                        (other_value is not None and other_value != []) and PREFER_OTHER
                    )
                    else self_value
                )

        # Recalculate doc_id if doi has changed
        if merged_data["doi"] != self.doi:
            merged_data["doc_id"] = (
                encode_id(merged_data["doi"].lower()) if merged_data["doi"] else None  # type: ignore[attr-defined,assignment]
            )

        # Create and return new DocDetails instance
        return DocDetails(**merged_data)

    def __radd__(self, other: DocDetails | int) -> DocDetails:
        # other == 0 captures the first call of sum()
        if isinstance(other, int) and other == 0:
            return self
        return self.__add__(other)

    def __iadd__(self, other: DocDetails | int) -> DocDetails:  # noqa: PYI034
        # only includes int to align with __radd__ and __add__
        if isinstance(other, int):
            return self
        return self.__add__(other)



from __future__ import annotations

import os
import shutil
from pathlib import Path
from typing import Generator
from unittest.mock import patch

import pytest
from dotenv import load_dotenv

from paperqa.clients.crossref import CROSSREF_HEADER_KEY
from paperqa.clients.semantic_scholar import SEMANTIC_SCHOLAR_HEADER_KEY
from paperqa.types import Answer


@pytest.fixture(autouse=True, scope="session")
def _load_env():
    load_dotenv()


@pytest.fixture(scope="session")
def vcr_config():
    return {
        "filter_headers": [CROSSREF_HEADER_KEY, SEMANTIC_SCHOLAR_HEADER_KEY],
    }


@pytest.fixture
def tmp_path_cleanup(
    tmp_path: Path,
) -> Generator[Path, None, None]:
    yield tmp_path
    # Cleanup after the test
    if tmp_path.exists():
        shutil.rmtree(tmp_path, ignore_errors=True)


@pytest.fixture
def agent_home_dir(
    tmp_path_cleanup: str | os.PathLike,
) -> Generator[str | os.PathLike, None, None]:
    """Set up a unique temporary folder for the agent module."""
    with patch.dict("os.environ", {"PQA_HOME": str(tmp_path_cleanup)}):
        yield tmp_path_cleanup


@pytest.fixture
def agent_index_dir(agent_home_dir: Path) -> Path:
    return agent_home_dir / ".pqa" / "indexes"


@pytest.fixture(name="agent_test_kit")
def fixture_stub_answer() -> Answer:
    return Answer(question="What is is a self-explanatory model?")


@pytest.fixture(name="stub_paper_path", scope="session")
def fixture_stub_paper_path() -> Path:
    # Corresponds with https://www.semanticscholar.org/paper/A-Perspective-on-Explanations-of-Molecular-Models-Wellawatte-Gandhi/1db1bde653658ec9b30858ae14650b8f9c9d438b
    return Path(__file__).parent / "paper.pdf"



from __future__ import annotations

from math import ceil
from pathlib import Path
from typing import Literal, overload

import html2text
import tiktoken

from .types import ChunkMetadata, Doc, ParsedMetadata, ParsedText, Text
from .version import __version__ as pqa_version


def parse_pdf_fitz_to_pages(path: Path) -> ParsedText:
    import fitz

    with fitz.open(path) as file:
        pages: dict[str, str] = {}
        total_length = 0

        for i in range(file.page_count):
            page = file.load_page(i)
            pages[str(i + 1)] = page.get_text("text", sort=True)
            total_length += len(pages[str(i + 1)])

    metadata = ParsedMetadata(
        parsing_libraries=[f"fitz ({fitz.__doc__})"],
        paperqa_version=str(pqa_version),
        total_parsed_text_length=total_length,
        parse_type="pdf",
    )
    return ParsedText(content=pages, metadata=metadata)


def parse_pdf_to_pages(path: Path) -> ParsedText:
    import pypdf

    with open(path, "rb") as pdfFileObj:
        pdfReader = pypdf.PdfReader(pdfFileObj)
        pages: dict[str, str] = {}
        total_length = 0

        for i, page in enumerate(pdfReader.pages):
            pages[str(i + 1)] = page.extract_text()
            total_length += len(pages[str(i + 1)])

    metadata = ParsedMetadata(
        parsing_libraries=[f"pypdf ({pypdf.__version__})"],
        paperqa_version=str(pqa_version),
        total_parsed_text_length=total_length,
        parse_type="pdf",
    )
    return ParsedText(content=pages, metadata=metadata)


def chunk_pdf(
    parsed_text: ParsedText, doc: Doc, chunk_chars: int, overlap: int
) -> list[Text]:
    pages: list[str] = []
    texts: list[Text] = []
    split: str = ""

    if not isinstance(parsed_text.content, dict):
        raise NotImplementedError(
            f"ParsedText.content must be a `dict`, not {type(parsed_text.content)}."
        )

    for page_num, page_text in parsed_text.content.items():
        split += page_text
        pages.append(page_num)
        # split could be so long it needs to be split
        # into multiple chunks. Or it could be so short
        # that it needs to be combined with the next chunk.
        while len(split) > chunk_chars:
            # pretty formatting of pages (e.g. 1-3, 4, 5-7)
            pg = "-".join([pages[0], pages[-1]])
            texts.append(
                Text(
                    text=split[:chunk_chars], name=f"{doc.docname} pages {pg}", doc=doc
                )
            )
            split = split[chunk_chars - overlap :]
            pages = [page_num]

    if len(split) > overlap or len(texts) == 0:
        pg = "-".join([pages[0], pages[-1]])
        texts.append(
            Text(text=split[:chunk_chars], name=f"{doc.docname} pages {pg}", doc=doc)
        )
    return texts


def parse_text(
    path: Path, html: bool = False, split_lines=False, use_tiktoken=True
) -> ParsedText:
    """Simple text splitter, can optionally use tiktoken, parse html, or split into newlines.

    Args:
        path: path to file
        html: flag to use html2text library for parsing
        split_lines: flag to split lines into a list
        use_tiktoken: flag to use tiktoken library to encode text

    """
    try:
        with open(path) as f:
            text = [str(line) for line in f] if split_lines else f.read()
    except UnicodeDecodeError:
        with open(path, encoding="utf-8", errors="ignore") as f:
            text = f.read()

    if html:
        text = html2text.html2text(text)

    metadata = {
        "parsing_libraries": ["tiktoken (cl100k_base)"] if use_tiktoken else [],
        "paperqa_version": str(pqa_version),
        "total_parsed_text_length": (
            len(text) if isinstance(text, str) else sum(len(t) for t in text)
        ),
        "parse_type": "txt" if not html else "html",
    }
    if html:
        metadata["parsing_libraries"].append(f"html2text ({html2text.__version__})")  # type: ignore[attr-defined]

    return ParsedText(content=text, metadata=ParsedMetadata(**metadata))


def chunk_text(
    parsed_text: ParsedText, doc: Doc, chunk_chars: int, overlap: int, use_tiktoken=True
) -> list[Text]:
    """Parse a document into chunks, based on tiktoken encoding.

    NOTE: We get some byte continuation errors.
    Currently ignored, but should explore more to make sure we
    don't miss anything.
    """
    texts: list[Text] = []
    enc = tiktoken.get_encoding("cl100k_base")
    split = []

    if not isinstance(parsed_text.content, str):
        raise NotImplementedError(
            f"ParsedText.content must be a `str`, not {type(parsed_text.content)}."
        )

    content = parsed_text.content if not use_tiktoken else parsed_text.encode_content()

    # convert from characters to chunks
    char_count = parsed_text.metadata.total_parsed_text_length  # e.g., 25,000
    token_count = len(content)  # e.g., 4,500
    chars_per_token = char_count / token_count  # e.g., 5.5
    chunk_tokens = chunk_chars / chars_per_token  # e.g., 3000 / 5.5 = 545
    overlap_tokens = overlap / chars_per_token  # e.g., 100 / 5.5 = 18
    chunk_count = ceil(token_count / chunk_tokens)  # e.g., 4500 / 545 = 9

    for i in range(chunk_count):
        split = content[
            max(int(i * chunk_tokens - overlap_tokens), 0) : int(
                (i + 1) * chunk_tokens + overlap_tokens
            )
        ]
        texts.append(
            Text(
                text=enc.decode(split) if use_tiktoken else split,
                name=f"{doc.docname} chunk {i + 1}",
                doc=doc,
            )
        )
    return texts


def chunk_code_text(
    parsed_text: ParsedText, doc: Doc, chunk_chars: int, overlap: int
) -> list[Text]:
    """Parse a document into chunks, based on line numbers (for code)."""
    split = ""
    texts: list[Text] = []
    last_line = 0

    if not isinstance(parsed_text.content, list):
        raise NotImplementedError(
            f"ParsedText.content must be a `list`, not {type(parsed_text.content)}."
        )

    for i, line in enumerate(parsed_text.content):
        split += line
        while len(split) > chunk_chars:
            texts.append(
                Text(
                    text=split[:chunk_chars],
                    name=f"{doc.docname} lines {last_line}-{i}",
                    doc=doc,
                )
            )
            split = split[chunk_chars - overlap :]
            last_line = i
    if len(split) > overlap or len(texts) == 0:
        texts.append(
            Text(
                text=split[:chunk_chars],
                name=f"{doc.docname} lines {last_line}-{i}",
                doc=doc,
            )
        )
    return texts


@overload
def read_doc(
    path: Path,
    doc: Doc,
    parsed_text_only: Literal[False],
    include_metadata: Literal[False],
    chunk_chars: int = ...,
    overlap: int = ...,
    force_pypdf: bool = ...,
) -> list[Text]: ...


@overload
def read_doc(
    path: Path,
    doc: Doc,
    parsed_text_only: Literal[False] = ...,
    include_metadata: Literal[False] = ...,
    chunk_chars: int = ...,
    overlap: int = ...,
    force_pypdf: bool = ...,
) -> list[Text]: ...


@overload
def read_doc(
    path: Path,
    doc: Doc,
    parsed_text_only: Literal[True],
    include_metadata: bool = ...,
    chunk_chars: int = ...,
    overlap: int = ...,
    force_pypdf: bool = ...,
) -> ParsedText: ...


@overload
def read_doc(
    path: Path,
    doc: Doc,
    parsed_text_only: Literal[False],
    include_metadata: Literal[True],
    chunk_chars: int = ...,
    overlap: int = ...,
    force_pypdf: bool = ...,
) -> tuple[list[Text], ParsedMetadata]: ...


def read_doc(
    path: Path,
    doc: Doc,
    parsed_text_only: bool = False,
    include_metadata: bool = False,
    chunk_chars: int = 3000,
    overlap: int = 100,
    force_pypdf: bool = False,
) -> list[Text] | ParsedText | tuple[list[Text], ParsedMetadata]:
    """Parse a document and split into chunks.

    Optionally can include just the parsing as well as metadata about the parsing/chunking

    Args:
        path: local document path
        doc: object with document metadata
        chunk_chars: size of chunks
        overlap: size of overlap between chunks
        force_pypdf: flag to force use of pypdf in parsing
        parsed_text_only: return parsed text without chunking
        include_metadata: return a tuple
    """
    str_path = str(path)
    parsed_text = None

    # start with parsing -- users may want to store this separately
    if str_path.endswith(".pdf"):
        if force_pypdf:
            parsed_text = parse_pdf_to_pages(path)
        else:
            try:
                parsed_text = parse_pdf_fitz_to_pages(path)
            except ImportError:
                parsed_text = parse_pdf_to_pages(path)

    elif str_path.endswith(".txt"):
        parsed_text = parse_text(path, html=False, split_lines=False, use_tiktoken=True)
    elif str_path.endswith(".html"):
        parsed_text = parse_text(path, html=True, split_lines=False, use_tiktoken=True)
    else:
        parsed_text = parse_text(path, html=False, split_lines=True, use_tiktoken=False)

    if parsed_text_only:
        return parsed_text

    # next chunk the parsed text
    if str_path.endswith(".pdf"):
        chunked_text = chunk_pdf(
            parsed_text, doc, chunk_chars=chunk_chars, overlap=overlap
        )
        chunk_metadata = ChunkMetadata(
            chunk_chars=chunk_chars, overlap=overlap, chunk_type="overlap_pdf_by_page"
        )
    elif str_path.endswith((".txt", ".html")):
        chunked_text = chunk_text(
            parsed_text,
            doc,
            chunk_chars=chunk_chars,
            overlap=overlap,
            use_tiktoken=True,
        )
        chunk_metadata = ChunkMetadata(
            chunk_chars=chunk_chars, overlap=overlap, chunk_type="overlap"
        )
    else:
        chunked_text = chunk_code_text(
            parsed_text, doc, chunk_chars=chunk_chars, overlap=overlap
        )
        chunk_metadata = ChunkMetadata(
            chunk_chars=chunk_chars, overlap=overlap, chunk_type="overlap_code_by_line"
        )

    if include_metadata:
        parsed_text.metadata.chunk_metadata = chunk_metadata
        return chunked_text, parsed_text.metadata

    return chunked_text



from __future__ import annotations

import copy
import json
import logging
import os
from datetime import datetime
from typing import Any, Collection
from urllib.parse import quote

import aiohttp

from ..clients.exceptions import DOINotFoundError
from ..types import CITATION_FALLBACK_DATA, DocDetails
from ..utils import (
    bibtex_field_extract,
    create_bibtex_key,
    remove_substrings,
    strings_similarity,
    union_collections_to_ordered_list,
)
from .client_models import DOIOrTitleBasedProvider, DOIQuery, TitleAuthorQuery

logger = logging.getLogger(__name__)

CROSSREF_BASE_URL = "https://api.crossref.org"
CROSSREF_HEADER_KEY = "Crossref-Plus-API-Token"
CROSSREF_API_REQUEST_TIMEOUT = 5.0
CROSSREF_API_MAPPING: dict[str, Collection[str]] = {
    "title": {"title"},
    "doi": {"DOI"},
    "authors": {"author"},
    "publication_date": {"published"},
    "year": {"published"},
    "volume": {"volume"},
    "issue": {"issue"},
    "publisher": {"publisher"},
    "issn": {"ISSN"},
    "pages": {"page"},
    "journal": {"container-title"},
    "doi_url": {"URL"},
    "url": {"URL"},
    "bibtex": {"bibtex", "type"},
    "citation_count": {"is-referenced-by-count"},
    "bibtex_type": {"type"},
    "citation": {
        "title",
        "DOI",
        "published",
        "volume",
        "issue",
        "publisher",
        "ISSN",
        "page",
        "container-title",
        "is-referenced-by-count",
        "type",
    },
    "source_quality": {"container-title"},
    "doc_id": {"DOI"},
}
CROSSREF_CONTENT_TYPE_TO_BIBTEX_MAPPING: dict[str, str] = {
    "journal-article": "article",
    "journal-issue": "misc",  # No direct equivalent, so 'misc' is used
    "journal-volume": "misc",  # No direct equivalent, so 'misc' is used
    "journal": "misc",  # No direct equivalent, so 'misc' is used
    "proceedings-article": "inproceedings",
    "proceedings": "proceedings",
    "dataset": "misc",  # No direct equivalent, so 'misc' is used
    "component": "misc",  # No direct equivalent, so 'misc' is used
    "report": "techreport",
    "report-series": "techreport",  # 'series' implies multiple tech reports, but each is still a 'techreport'
    "standard": "misc",  # No direct equivalent, so 'misc' is used
    "standard-series": "misc",  # No direct equivalent, so 'misc' is used
    "edited-book": "book",  # Edited books are considered books in BibTeX
    "monograph": "book",  # Monographs are considered books in BibTeX
    "reference-book": "book",  # Reference books are considered books in BibTeX
    "book": "book",
    "book-series": "book",  # Series of books can be considered as 'book' in BibTeX
    "book-set": "book",  # Set of books can be considered as 'book' in BibTeX
    "book-chapter": "inbook",
    "book-section": "inbook",  # Sections in books can be considered as 'inbook'
    "book-part": "inbook",  # Parts of books can be considered as 'inbook'
    "book-track": "inbook",  # Tracks in books can be considered as 'inbook'
    "reference-entry": "inbook",  # Entries in reference books can be considered as 'inbook'
    "dissertation": "phdthesis",  # Dissertations are usually PhD thesis
    "posted-content": "misc",  # No direct equivalent, so 'misc' is used
    "peer-review": "misc",  # No direct equivalent, so 'misc' is used
    "other": "article",  # Assume an article if we don't know the type
}


def crossref_headers() -> dict[str, str]:
    """Crossref API key if available, otherwise nothing."""
    if api_key := os.environ.get("CROSSREF_API_KEY"):
        return {CROSSREF_HEADER_KEY: f"Bearer {api_key}"}
    logger.warning(
        "CROSSREF_API_KEY environment variable not set. Crossref API rate limits may apply."
    )
    return {}


async def doi_to_bibtex(
    doi: str,
    session: aiohttp.ClientSession,
    missing_replacements: dict[str, str] | None = None,
) -> str:
    """Get a bibtex entry from a DOI via Crossref, replacing the key if possible.

    `missing_replacements` can optionally be used to fill missing fields in the bibtex key.
        these fields are NOT replaced or inserted into the bibtex string otherwise.

    """
    if missing_replacements is None:
        missing_replacements = {}
    FORBIDDEN_KEY_CHARACTERS = {"_", " ", "-", "/"}
    # get DOI via crossref
    url = f"https://api.crossref.org/works/{quote(doi, safe='')}/transform/application/x-bibtex"
    async with session.get(url, headers=crossref_headers()) as r:
        if not r.ok:
            raise DOINotFoundError(
                f"Per HTTP status code {r.status}, could not resolve DOI {doi}."
            )
        data = await r.text()
    # must make new key
    key = data.split("{")[1].split(",")[0]
    new_key = remove_substrings(key, FORBIDDEN_KEY_CHARACTERS)
    substrings_to_remove_per_field = {"author": [" and ", ","]}
    fragments = [
        remove_substrings(
            bibtex_field_extract(
                data, field, missing_replacements=missing_replacements
            ),
            substrings_to_remove_per_field.get(field, []),
        )
        for field in ("author", "year", "title")
    ]
    # replace the key if all the fragments are present
    if all(fragments):
        new_key = create_bibtex_key(
            author=fragments[0].split(), year=fragments[1], title=fragments[2]
        )
    # we use the count parameter below to ensure only the 1st entry is replaced
    return data.replace(key, new_key, 1)


async def parse_crossref_to_doc_details(  # noqa: C901
    message: dict[str, Any],
    session: aiohttp.ClientSession,
    query_bibtex: bool = True,
) -> DocDetails:

    bibtex_source = "self_generated"
    bibtex = None

    try:
        # get the title from the message, if it exists
        # rare circumstance, but bibtex may not have a title
        fallback_data = copy.copy(CITATION_FALLBACK_DATA)
        if title := (
            None if not message.get("title") else message.get("title", [None])[0]
        ):
            fallback_data["title"] = title

        # TODO: we keep this for robustness, but likely not needed anymore,
        # since we now create the bibtex from scratch
        if query_bibtex:
            bibtex = await doi_to_bibtex(
                message["DOI"], session, missing_replacements=fallback_data  # type: ignore[arg-type]
            )
            # track the origin of the bibtex entry for debugging
            bibtex_source = "crossref"

    except DOINotFoundError:
        pass

    authors = [
        f"{author.get('given', '')} {author.get('family', '')}".strip()
        for author in message.get("author", [])
    ]

    publication_date = None
    if "published" in message and "date-parts" in message["published"]:
        date_parts = message["published"]["date-parts"][0]
        if len(date_parts) >= 3:  # noqa: PLR2004
            publication_date = datetime(date_parts[0], date_parts[1], date_parts[2])
        elif len(date_parts) == 2:  # noqa: PLR2004
            publication_date = datetime(date_parts[0], date_parts[1], 1)
        elif len(date_parts) == 1:
            publication_date = datetime(date_parts[0], 1, 1)

    doc_details = DocDetails(  # type: ignore[call-arg]
        key=None if not bibtex else bibtex.split("{")[1].split(",")[0],
        bibtex_type=CROSSREF_CONTENT_TYPE_TO_BIBTEX_MAPPING.get(
            message.get("type", "other"), "misc"
        ),
        bibtex=bibtex,
        authors=authors,
        publication_date=publication_date,
        year=message.get("published", {}).get("date-parts", [[None]])[0][0],
        volume=message.get("volume"),
        issue=message.get("issue"),
        publisher=message.get("publisher"),
        issn=message.get("ISSN", [None])[0],
        pages=message.get("page"),
        journal=(
            None
            if not message.get("container-title")
            else message["container-title"][0]
        ),
        url=message.get("URL"),
        title=None if not message.get("title") else message.get("title", [None])[0],
        citation_count=message.get("is-referenced-by-count"),
        doi=message.get("DOI"),
        other={},  # Initialize empty dict for other fields
    )

    # Add any additional fields to the 'other' dict
    for key, value in (
        message | {"client_source": ["crossref"], "bibtex_source": [bibtex_source]}
    ).items():
        if key not in doc_details.model_fields:
            if key in doc_details.other:
                doc_details.other[key] = [doc_details.other[key], value]
            else:
                doc_details.other[key] = value

    return doc_details


async def get_doc_details_from_crossref(  # noqa: C901, PLR0912
    session: aiohttp.ClientSession,
    doi: str | None = None,
    authors: list[str] | None = None,
    title: str | None = None,
    title_similarity_threshold: float = 0.75,
    fields: Collection[str] | None = None,
) -> DocDetails | None:
    """
    Get paper details from Crossref given a DOI or paper title.

    SEE: https://api.crossref.org/swagger-ui/index.html#/Works
    """
    if authors is None:
        authors = []
    if doi is title is None:
        raise ValueError("Either a DOI or title must be provided.")
    if doi is not None and title is not None:
        title = None  # Prefer DOI over title

    inputs_msg = f"DOI {doi}" if doi is not None else f"title {title}"

    if not (CROSSREF_MAILTO := os.getenv("CROSSREF_MAILTO")):
        logger.warning(
            "CROSSREF_MAILTO environment variable not set. Crossref API rate limits may apply."
        )
        CROSSREF_MAILTO = "test@example.com"
    quoted_doi = f"/{quote(doi, safe='')}" if doi else ""
    url = f"{CROSSREF_BASE_URL}/works{quoted_doi}"
    params = {"mailto": CROSSREF_MAILTO}
    if title:
        params.update({"query.title": title, "rows": "1"})

    if authors:
        params.update(
            {"query.author": " ".join([a.strip() for a in authors if len(a) > 1])}
        )

    query_bibtex = True

    # note we only do field selection if querying on title
    if fields and title:
        # crossref has a special endpoint for bibtex, so we don't need to request it here
        if "bibtex" not in fields:
            query_bibtex = False
        params.update(
            {
                "select": ",".join(
                    union_collections_to_ordered_list(
                        [
                            CROSSREF_API_MAPPING[field]
                            for field in fields
                            if field in CROSSREF_API_MAPPING and field != "bibtex"
                        ]
                    )
                )
            }
        )

    async with session.get(
        url,
        params=params,
        headers=crossref_headers(),
        timeout=aiohttp.ClientTimeout(CROSSREF_API_REQUEST_TIMEOUT),
    ) as response:
        try:
            response.raise_for_status()
        except aiohttp.ClientResponseError as exc:
            raise DOINotFoundError(f"Could not find paper given {inputs_msg}.") from exc
        try:
            response_data = await response.json()
        except json.JSONDecodeError as exc:
            # JSONDecodeError: Crossref didn't answer with JSON, perhaps HTML
            raise DOINotFoundError(  # Use DOINotFoundError so we fall back to Google Scholar
                f"Crossref API did not return JSON for {inputs_msg}, instead it"
                f" responded with text: {await response.text()}"
            ) from exc
    if response_data["status"] == "failed":
        raise DOINotFoundError(
            f"Crossref API returned a failed status for {inputs_msg}."
        )
    message: dict[str, Any] = response_data["message"]
    # restructure data if it comes back as a list result
    # it'll also be a list if we searched by title and it's empty
    if "items" in message:
        try:
            message = message["items"][0]
        except IndexError as e:
            raise DOINotFoundError(
                f"Crossref API did not return any items for {inputs_msg}."
            ) from e
    # since score is not consistent between queries, we need to rely on our own criteria
    # title similarity must be > title_similarity_threshold
    if (
        doi is None
        and title
        and strings_similarity(message["title"][0], title) < title_similarity_threshold
    ):
        raise DOINotFoundError(f"Crossref results did not match for title {title!r}.")
    if doi is not None and message["DOI"] != doi:
        raise DOINotFoundError(f"DOI ({inputs_msg}) not found in Crossref")

    return await parse_crossref_to_doc_details(message, session, query_bibtex)


class CrossrefProvider(DOIOrTitleBasedProvider):
    async def _query(self, query: TitleAuthorQuery | DOIQuery) -> DocDetails | None:
        if isinstance(query, DOIQuery):
            return await get_doc_details_from_crossref(
                doi=query.doi, session=query.session, fields=query.fields
            )
        return await get_doc_details_from_crossref(
            title=query.title,
            authors=query.authors,
            session=query.session,
            title_similarity_threshold=query.title_similarity_threshold,
            fields=query.fields,
        )



from __future__ import annotations

import logging
import os
import re
from datetime import datetime
from typing import Any, cast

from anthropic import AsyncAnthropic
from openai import AsyncOpenAI
from rich.table import Table

from .. import (
    AnthropicLLMModel,
    Docs,
    OpenAILLMModel,
    embedding_model_factory,
    llm_model_factory,
)
from ..llms import LangchainLLMModel
from .models import AnswerResponse, QueryRequest

logger = logging.getLogger(__name__)


def get_year(ts: datetime | None = None) -> str:
    """Get the year from the input datetime, otherwise using the current datetime."""
    if ts is None:
        ts = datetime.now()
    return ts.strftime("%Y")


async def openai_get_search_query(
    question: str,
    count: int,
    template: str | None = None,
    llm: str = "gpt-4o-mini",
    temperature: float = 1.0,
) -> list[str]:
    if isinstance(template, str):
        if not (
            "{count}" in template and "{question}" in template and "{date}" in template
        ):
            logger.warning(
                "Template does not contain {count}, {question} and {date} variables. Ignoring template."
            )
            template = None

        else:
            # partial formatting
            search_prompt = template.replace("{date}", get_year())

    if template is None:
        search_prompt = (
            "We want to answer the following question: {question} \n"
            "Provide {count} unique keyword searches (one search per line) and year ranges "
            "that will find papers to help answer the question. "
            "Do not use boolean operators. "
            "Make sure not to repeat searches without changing the keywords or year ranges. "
            "Make some searches broad and some narrow. "
            "Use this format: [keyword search], [start year]-[end year]. "
            "where end year is optional. "
            f"The current year is {get_year()}."
        )

    if "gpt" not in llm:
        raise ValueError(
            f"Invalid llm: {llm}, note a GPT model must be used for the fake agent search."
        )
    client = AsyncOpenAI()
    model = OpenAILLMModel(config={"model": llm, "temperature": temperature})
    chain = model.make_chain(client, prompt=search_prompt, skip_system=True)
    result = await chain({"question": question, "count": count})  # type: ignore[call-arg]
    search_query = result.text
    queries = [s for s in search_query.split("\n") if len(s) > 3]  # noqa: PLR2004
    # remove "2.", "3.", etc. -- https://regex101.com/r/W2f7F1/1
    queries = [re.sub(r"^\d+\.\s*", "", q) for q in queries]
    # remove quotes
    return [re.sub(r"\"", "", q) for q in queries]


def table_formatter(
    objects: list[tuple[AnswerResponse | Docs, str]], max_chars_per_column: int = 2000
) -> Table:
    example_object, _ = objects[0]
    if isinstance(example_object, AnswerResponse):
        table = Table(title="Prior Answers")
        table.add_column("Question", style="cyan")
        table.add_column("Answer", style="magenta")
        for obj, _ in objects:
            table.add_row(
                cast(AnswerResponse, obj).answer.question[:max_chars_per_column],
                cast(AnswerResponse, obj).answer.answer[:max_chars_per_column],
            )
        return table
    if isinstance(example_object, Docs):
        table = Table(title="PDF Search")
        table.add_column("Title", style="cyan")
        table.add_column("File", style="magenta")
        for obj, filename in objects:
            table.add_row(
                cast(Docs, obj).texts[0].doc.title[:max_chars_per_column], filename  # type: ignore[attr-defined]
            )
        return table
    raise NotImplementedError(
        f"Object type {type(example_object)} can not be converted to table."
    )


# Index 0 is for prompt tokens, index 1 is for completion tokens
costs: dict[str, tuple[float, float]] = {
    "claude-2": (11.02 / 10**6, 32.68 / 10**6),
    "claude-instant-1": (1.63 / 10**6, 5.51 / 10**6),
    "claude-3-sonnet-20240229": (3 / 10**6, 15 / 10**6),
    "claude-3-5-sonnet-20240620": (3 / 10**6, 15 / 10**6),
    "claude-3-opus-20240229": (15 / 10**6, 75 / 10**6),
    "babbage-002": (0.0004 / 10**3, 0.0004 / 10**3),
    "gpt-3.5-turbo": (0.0010 / 10**3, 0.0020 / 10**3),
    "gpt-3.5-turbo-1106": (0.0010 / 10**3, 0.0020 / 10**3),
    "gpt-3.5-turbo-0613": (0.0010 / 10**3, 0.0020 / 10**3),
    "gpt-3.5-turbo-0301": (0.0010 / 10**3, 0.0020 / 10**3),
    "gpt-3.5-turbo-0125": (0.0005 / 10**3, 0.0015 / 10**3),
    "gpt-4-1106-preview": (0.010 / 10**3, 0.030 / 10**3),
    "gpt-4-0125-preview": (0.010 / 10**3, 0.030 / 10**3),
    "gpt-4-turbo-2024-04-09": (10 / 10**6, 30 / 10**6),
    "gpt-4-turbo": (10 / 10**6, 30 / 10**6),
    "gpt-4": (0.03 / 10**3, 0.06 / 10**3),
    "gpt-4-0613": (0.03 / 10**3, 0.06 / 10**3),
    "gpt-4-0314": (0.03 / 10**3, 0.06 / 10**3),
    "gpt-4o": (2.5 / 10**6, 10 / 10**6),
    "gpt-4o-2024-05-13": (5 / 10**6, 15 / 10**6),
    "gpt-4o-2024-08-06": (2.5 / 10**6, 10 / 10**6),
    "gpt-4o-mini": (0.15 / 10**6, 0.60 / 10**6),
    "gemini-1.5-flash": (0.35 / 10**6, 0.35 / 10**6),
    "gemini-1.5-pro": (3.5 / 10**6, 10.5 / 10**6),
    # supported Anyscale models per
    # https://docs.anyscale.com/endpoints/text-generation/query-a-model
    "meta-llama/Meta-Llama-3-8B-Instruct": (0.15 / 10**6, 0.15 / 10**6),
    "meta-llama/Meta-Llama-3-70B-Instruct": (1.0 / 10**6, 1.0 / 10**6),
    "mistralai/Mistral-7B-Instruct-v0.1": (0.15 / 10**6, 0.15 / 10**6),
    "mistralai/Mixtral-8x7B-Instruct-v0.1": (1.0 / 10**6, 1.0 / 10**6),
    "mistralai/Mixtral-8x22B-Instruct-v0.1": (1.0 / 10**6, 1.0 / 10**6),
}


def compute_model_token_cost(model: str, tokens: int, is_completion: bool) -> float:
    if model in costs:  # Prefer our internal costs model
        model_costs: tuple[float, float] = costs[model]
    else:
        logger.warning(f"Model {model} not found in costs.")
        return 0.0
    return tokens * model_costs[int(is_completion)]


def compute_total_model_token_cost(token_counts: dict[str, list[int]]) -> float:
    """Sum the token counts for each model and return the total cost."""
    cost = 0.0
    for model, tokens in token_counts.items():
        if sum(tokens) > 0:
            cost += compute_model_token_cost(
                model, tokens=tokens[0], is_completion=False
            ) + compute_model_token_cost(model, tokens=tokens[1], is_completion=True)
    return cost


# the defaults here should be (about) the same as in QueryRequest
def update_doc_models(doc: Docs, request: QueryRequest | None = None):
    if request is None:
        request = QueryRequest()
    client: Any = None

    if request.llm.startswith("gemini"):
        doc.llm_model = LangchainLLMModel(name=request.llm)
        doc.summary_llm_model = LangchainLLMModel(name=request.summary_llm)
    else:
        doc.llm_model = llm_model_factory(request.llm)
        doc.summary_llm_model = llm_model_factory(request.summary_llm)

    # set temperatures
    doc.llm_model.config["temperature"] = request.temperature
    doc.summary_llm_model.config["temperature"] = request.temperature

    if isinstance(doc.llm_model, OpenAILLMModel):
        if request.llm.startswith(
            ("meta-llama/Meta-Llama-3-", "mistralai/Mistral-", "mistralai/Mixtral-")
        ):
            client = AsyncOpenAI(
                base_url=os.environ.get("ANYSCALE_BASE_URL"),
                api_key=os.environ.get("ANYSCALE_API_KEY"),
            )
            logger.info(f"Using Anyscale (via OpenAI client) for {request.llm}")
        else:
            client = AsyncOpenAI()
    elif isinstance(doc.llm_model, AnthropicLLMModel):
        client = AsyncAnthropic()
    elif isinstance(doc.llm_model, LangchainLLMModel):
        from langchain_google_vertexai import (
            ChatVertexAI,
            HarmBlockThreshold,
            HarmCategory,
        )

        # we have to convert system to human because system is unsupported
        # Also we do get blocked content, so adjust thresholds
        client = ChatVertexAI(
            model=request.llm,
            safety_settings={
                HarmCategory.HARM_CATEGORY_DANGEROUS_CONTENT: HarmBlockThreshold.BLOCK_ONLY_HIGH,
                HarmCategory.HARM_CATEGORY_HATE_SPEECH: HarmBlockThreshold.BLOCK_ONLY_HIGH,
                HarmCategory.HARM_CATEGORY_HARASSMENT: HarmBlockThreshold.BLOCK_ONLY_HIGH,
                HarmCategory.HARM_CATEGORY_SEXUALLY_EXPLICIT: HarmBlockThreshold.BLOCK_ONLY_HIGH,
            },
            convert_system_message_to_human=True,
        )
    else:
        raise TypeError(f"Unsupported LLM model: {doc.llm_model}")

    doc._client = client  # set client, since could be just unpickled.
    doc._embedding_client = AsyncOpenAI()  # hard coded to OpenAI for now

    doc.texts_index.embedding_model = embedding_model_factory(
        request.embedding, **(request.texts_index_embedding_config or {})
    )
    doc.docs_index.embedding_model = embedding_model_factory(
        request.embedding, **(request.docs_index_embedding_config or {})
    )
    doc.texts_index.mmr_lambda = request.texts_index_mmr_lambda
    doc.docs_index.mmr_lambda = request.docs_index_mmr_lambda
    doc.embedding = request.embedding
    doc.max_concurrent = request.max_concurrent
    doc.prompts = request.prompts
    Docs.make_llm_names_consistent(doc)

    logger.debug(
        f"update_doc_models: {doc.name}"
        f" | {(doc.llm_model.config)} | {(doc.summary_llm_model.config)}"
        f" | {doc.docs_index.__class__}"
    )



from __future__ import annotations

import logging
from typing import Any, Collection, Sequence, cast

import aiohttp
import pytest

import paperqa
from paperqa.clients import (
    CrossrefProvider,
    DocMetadataClient,
    SemanticScholarProvider,
)
from paperqa.clients.client_models import MetadataPostProcessor, MetadataProvider
from paperqa.clients.journal_quality import JournalQualityPostProcessor


@pytest.mark.vcr()
@pytest.mark.parametrize(
    ("paper_attributes"),
    [
        {
            "title": (
                "Effect of native oxide layers on copper thin-film "
                "tensile properties: A reactive molecular dynamics study"
            ),
            "source": ["semantic_scholar", "crossref"],
            "key": "skarlinski2015effectofnative",
            "doi": "10.1063/1.4938384",
            "doc_id": "c217ec9289696c3c",
            "journal": "Journal of Applied Physics",
            "authors": ["Michael D. Skarlinski", "David J. Quesnel"],
            "formatted_citation": (
                "Michael D. Skarlinski and David J. Quesnel. Effect of native "
                "oxide layers on copper thin-film tensile properties: a reactive"
                " molecular dynamics study. Journal of Applied Physics, 118:235306, "
                "Dec 2015. URL: https://doi.org/10.1063/1.4938384, doi:10.1063/1.4938384. "
                "This article has 8 citations and is from a peer-reviewed journal."
            ),
        },
        {
            "title": "PaperQA: Retrieval-Augmented Generative Agent for Scientific Research",
            "source": ["semantic_scholar"],
            "key": "lala2023paperqaretrievalaugmentedgenerative",
            "doi": "10.48550/arxiv.2312.07559",
            "doc_id": "bb985e0e3265d678",
            "journal": "ArXiv",
            "authors": [
                "Jakub L'ala",
                "Odhran O'Donoghue",
                "Aleksandar Shtedritski",
                "Sam Cox",
                "Samuel G. Rodriques",
                "Andrew D. White",
            ],
            "formatted_citation": (
                "Jakub L'ala, Odhran O'Donoghue, Aleksandar Shtedritski, Sam Cox, Samuel G. Rodriques,"
                " and Andrew D. White. Paperqa: retrieval-augmented generative agent for scientific "
                "research. ArXiv, Dec 2023. URL: https://doi.org/10.48550/arxiv.2312.07559, "
                "doi:10.48550/arxiv.2312.07559. This article has 21 citations."
            ),
        },
        {
            "title": "Augmenting large language models with chemistry tools",
            "source": ["semantic_scholar", "crossref"],
            "key": "bran2024augmentinglargelanguage",
            "doi": "10.1038/s42256-024-00832-8",
            "doc_id": "0f650d59b0a2ba5a",
            "journal": "Nature Machine Intelligence",
            "authors": [
                "Andres M. Bran",
                "Sam Cox",
                "Oliver Schilter",
                "Carlo Baldassari",
                "Andrew D. White",
                "Philippe Schwaller",
            ],
            "formatted_citation": (
                "Andres M. Bran, Sam Cox, Oliver Schilter, Carlo Baldassari, Andrew D. White, and "
                "Philippe Schwaller. Augmenting large language models with chemistry tools. Nature "
                "Machine Intelligence, 6:525-535, May 2024. URL: https://doi.org/10.1038/s42256-024-00832-8, "
                "doi:10.1038/s42256-024-00832-8. This article has 187 citations and is from a "
                "domain leading peer-reviewed journal."
            ),
        },
    ],
)
@pytest.mark.asyncio()
async def test_title_search(paper_attributes: dict[str, str]):
    async with aiohttp.ClientSession() as session:
        client = DocMetadataClient(session)
        details = await client.query(title=paper_attributes["title"])
        assert set(details.other["client_source"]) == set(  # type: ignore[union-attr]
            paper_attributes["source"]
        ), "Should have the correct source"
        for key, value in paper_attributes.items():
            if key != "source":
                assert getattr(details, key) == value, f"Should have the correct {key}"


@pytest.mark.vcr()
@pytest.mark.parametrize(
    ("paper_attributes"),
    [
        {
            "title": "High-throughput screening of human genetic variants by pooled prime editing",
            "source": ["semantic_scholar", "crossref"],
            "key": "herger2024highthroughputscreeningof",
            "doi": "10.1101/2024.04.01.587366",
            "doc_id": "8e7669b50f31c52b",
            "journal": "bioRxiv",
            "authors": [
                "Michael Herger",
                "Christina M. Kajba",
                "Megan Buckley",
                "Ana Cunha",
                "Molly Strom",
                "Gregory M. Findlay",
            ],
            "formatted_citation": (
                "Michael Herger, Christina M. Kajba, Megan Buckley, Ana Cunha, Molly Strom, "
                "and Gregory M. Findlay. High-throughput screening of human genetic variants "
                "by pooled prime editing. bioRxiv, Apr 2024. URL: https://doi.org/10.1101/2024.04.01.587366, "
                "doi:10.1101/2024.04.01.587366. This article has 1 citations."
            ),
        },
        {
            "title": (
                "An essential role of active site arginine residue in iodide binding and histidine residue "
                "in electron transfer for iodide oxidation by horseradish peroxidase"
            ),
            "source": ["semantic_scholar", "crossref"],
            "key": "adak2001anessentialrole",
            "doi": "10.1023/a:1007154515475",
            "doc_id": "3012c6676b658a27",
            "journal": "Molecular and Cellular Biochemistry",
            "authors": [
                "Subrata Adak",
                "Debashis Bandyopadhyay",
                "Uday Bandyopadhyay",
                "Ranajit K. Banerjee",
            ],
            "formatted_citation": (
                "Subrata Adak, Debashis Bandyopadhyay, Uday Bandyopadhyay, and Ranajit K. Banerjee. "
                "An essential role of active site arginine residue in iodide binding and histidine residue "
                "in electron transfer for iodide oxidation by horseradish peroxidase. Molecular and Cellular "
                "Biochemistry, 218:1-11, Feb 2001. URL: https://doi.org/10.1023/a:1007154515475, "
                "doi:10.1023/a:1007154515475. This article has 7 citations and is from a peer-reviewed journal."
            ),
        },
        {
            "title": "Convalescent-anti-sars-cov-2-plasma/immune-globulin",
            "source": ["semantic_scholar", "crossref"],
            "key": "unknownauthors2023convalescentantisarscov2plasmaimmuneglobulin",
            "doi": "10.1007/s40278-023-41815-2",
            "doc_id": "c2a60b772778732c",
            "journal": "Reactions Weekly",
            "authors": [],
            "formatted_citation": (
                "Unknown author(s). Convalescent-anti-sars-cov-2-plasma/immune-globulin. Reactions Weekly, "
                "1962:145-145, Jun 2023. URL: https://doi.org/10.1007/s40278-023-41815-2, "
                "doi:10.1007/s40278-023-41815-2. This article has 0 citations and is from a peer-reviewed journal."
            ),
        },
    ],
)
@pytest.mark.asyncio()
async def test_doi_search(paper_attributes: dict[str, str]):
    async with aiohttp.ClientSession() as session:
        client = DocMetadataClient(session)
        details = await client.query(doi=paper_attributes["doi"])
        assert set(details.other["client_source"]) == set(  # type: ignore[union-attr]
            paper_attributes["source"]
        ), "Should have the correct source"
        for key, value in paper_attributes.items():
            if key != "source":
                assert getattr(details, key) == value, f"Should have the correct {key}"


@pytest.mark.vcr()
@pytest.mark.asyncio()
async def test_bulk_doi_search():
    dois = [
        "10.1063/1.4938384",
        "10.48550/arxiv.2312.07559",
        "10.1038/s42256-024-00832-8",
        "10.1101/2024.04.01.587366",
        "10.1023/a:1007154515475",
        "10.1007/s40278-023-41815-2",
    ]
    async with aiohttp.ClientSession() as session:
        client = DocMetadataClient(session)
        details = await client.bulk_query([{"doi": doi} for doi in dois])
        assert len(details) == 6, "Should return 6 results"
        assert all(d for d in details), "All results should be non-None"


@pytest.mark.vcr()
@pytest.mark.asyncio()
async def test_bulk_title_search():
    titles = [
        "Effect of native oxide layers on copper thin-film tensile properties: A reactive molecular dynamics study",
        "PaperQA: Retrieval-Augmented Generative Agent for Scientific Research",
        "Augmenting large language models with chemistry tools",
        "High-throughput screening of human genetic variants by pooled prime editing",
        (
            "An essential role of active site arginine residue in iodide binding and histidine residue "
            "in electron transfer for iodide oxidation by horseradish peroxidase"
        ),
        "Convalescent-anti-sars-cov-2-plasma/immune-globulin",
    ]
    async with aiohttp.ClientSession() as session:
        client = DocMetadataClient(session)
        details = await client.bulk_query([{"title": title} for title in titles])
        assert len(details) == 6, "Should return 6 results"
        assert all(d for d in details), "All results should be non-None"


@pytest.mark.vcr()
@pytest.mark.asyncio()
async def test_bad_titles():
    async with aiohttp.ClientSession() as session:
        client = DocMetadataClient(session)
        details = await client.query(title="askldjrq3rjaw938h")
        assert not details, "Should return None for bad title"
        details = await client.query(
            title="Effect of native oxide layers on copper thin-film tensile properties: A study"
        )
        assert details, "Should find a similar title"


@pytest.mark.vcr()
@pytest.mark.asyncio()
async def test_bad_dois():
    async with aiohttp.ClientSession() as session:
        client = DocMetadataClient(session)
        details = await client.query(title="abs12032jsdafn")
        assert not details, "Should return None for bad doi"


@pytest.mark.vcr()
@pytest.mark.asyncio()
async def test_minimal_fields_filtering():
    async with aiohttp.ClientSession() as session:
        client = DocMetadataClient(session)
        details = await client.query(
            title="Augmenting large language models with chemistry tools",
            fields=["title", "doi"],
        )
        assert not details.journal, "Journal should not be populated"  # type: ignore[union-attr]
        assert not details.year, "Year should not be populated"  # type: ignore[union-attr]
        assert not details.authors, "Authors should not be populated"  # type: ignore[union-attr]
        assert details.citation == (  # type: ignore[union-attr]
            "Unknown author(s). Augmenting large language models with chemistry tools."
            " Unknown journal, Unknown year. URL: https://doi.org/10.1038/s42256-024-00832-8, "
            "doi:10.1038/s42256-024-00832-8."
        ), "Citation should be populated"
        assert set(details.other["client_source"]) == {  # type: ignore[union-attr]
            "semantic_scholar",
            "crossref",
        }, "Should be from two sources"
        assert not details.source_quality, "No source quality data should exist"  # type: ignore[union-attr]


@pytest.mark.vcr()
@pytest.mark.asyncio()
async def test_s2_only_fields_filtering():
    async with aiohttp.ClientSession() as session:
        # now get with authors just from one source
        s2_client = DocMetadataClient(session, clients=[SemanticScholarProvider])
        s2_details = await s2_client.query(
            title="Augmenting large language models with chemistry tools",
            fields=["title", "doi", "authors"],
        )
        assert s2_details.authors, "Authors should be populated"  # type: ignore[union-attr]
        assert set(s2_details.other["client_source"]) == {"semantic_scholar"}  # type: ignore[union-attr]
        assert s2_details.citation == (  # type: ignore[union-attr]
            "Andrés M Bran, Sam Cox, Oliver Schilter, Carlo Baldassari, Andrew D. White, "
            "and P. Schwaller. Augmenting large language models with chemistry tools. "
            "Unknown journal, Unknown year. URL: https://doi.org/10.1038/s42256-024-00832-8, "
            "doi:10.1038/s42256-024-00832-8."
        ), "Citation should be populated"
        assert not s2_details.source_quality, "No source quality data should exist"  # type: ignore[union-attr]


@pytest.mark.vcr()
@pytest.mark.asyncio()
async def test_crossref_journalquality_fields_filtering():
    async with aiohttp.ClientSession() as session:
        crossref_client = DocMetadataClient(
            session,
            clients=cast(
                Collection[
                    type[MetadataPostProcessor[Any]] | type[MetadataProvider[Any]]
                ],
                [CrossrefProvider, JournalQualityPostProcessor],
            ),
        )
        crossref_details = await crossref_client.query(
            title="Augmenting large language models with chemistry tools",
            fields=["title", "doi", "authors", "journal"],
        )
        assert set(crossref_details.other["client_source"]) == {  # type: ignore[union-attr]
            "crossref"
        }, "Should be from only crossref"
        assert crossref_details.source_quality == 2, "Should have source quality data"  # type: ignore[union-attr]
        assert crossref_details.citation == (  # type: ignore[union-attr]
            "Andres M. Bran, Sam Cox, Oliver Schilter, Carlo Baldassari, Andrew D. White, "
            "and Philippe Schwaller. Augmenting large language models with chemistry tools. "
            "Nature Machine Intelligence, Unknown year. URL: https://doi.org/10.1038/s42256-024-00832-8, "
            "doi:10.1038/s42256-024-00832-8."
        ), "Citation should be populated"


@pytest.mark.vcr()
@pytest.mark.asyncio()
async def test_author_matching():
    async with aiohttp.ClientSession() as session:
        crossref_client = DocMetadataClient(session, clients=[CrossrefProvider])
        s2_client = DocMetadataClient(session, clients=[SemanticScholarProvider])
        crossref_details_bad_author = await crossref_client.query(
            title="Augmenting large language models with chemistry tools",
            authors=["Jack NoScience"],
            fields=["title", "doi", "authors"],
        )

        s2_details_bad_author = await s2_client.query(
            title="Augmenting large language models with chemistry tools",
            authors=["Jack NoScience"],
            fields=["title", "doi", "authors"],
        )

        s2_details_w_author = await s2_client.query(
            title="Augmenting large language models with chemistry tools",
            authors=["Andres M. Bran", "Sam Cox"],
            fields=["title", "doi", "authors"],
        )

        assert not crossref_details_bad_author, "Should return None for bad author"
        assert not s2_details_bad_author, "Should return None for bad author"
        assert s2_details_w_author, "Should return results for good author"


@pytest.mark.vcr()
@pytest.mark.asyncio()
async def test_odd_client_requests():
    # try querying using an authors match, but not requesting authors back
    async with aiohttp.ClientSession() as session:
        client = DocMetadataClient(session)
        details = await client.query(
            title="Augmenting large language models with chemistry tools",
            authors=["Andres M. Bran", "Sam Cox"],
            fields=["title", "doi"],
        )
        assert (
            details.authors  # type: ignore[union-attr]
        ), "Should return correct author results"

    # try querying using a title, asking for no DOI back
    async with aiohttp.ClientSession() as session:
        client = DocMetadataClient(session)
        details = await client.query(
            title="Augmenting large language models with chemistry tools",
            fields=["title"],
        )
        assert (
            details.doi  # type: ignore[union-attr]
        ), "Should return a doi even though we don't ask for it"

    # try querying using a title, asking for no title back
    async with aiohttp.ClientSession() as session:
        client = DocMetadataClient(session)
        details = await client.query(
            title="Augmenting large language models with chemistry tools",
            fields=["doi"],
        )
        assert (
            details.title  # type: ignore[union-attr]
        ), "Should return a title even though we don't ask for it"

    async with aiohttp.ClientSession() as session:
        client = DocMetadataClient(session)
        details = await client.query(
            doi="10.1007/s40278-023-41815-2",
            fields=["doi", "title", "gibberish-field", "no-field"],
        )
        assert (
            details.title  # type: ignore[union-attr]
        ), "Should return title even though we asked for some bad fields"


@pytest.mark.asyncio()
async def test_ensure_robust_to_timeouts(monkeypatch):
    # 0.15 should be short enough to not get a response in time.
    monkeypatch.setattr(paperqa.clients.crossref, "CROSSREF_API_REQUEST_TIMEOUT", 0.05)
    monkeypatch.setattr(
        paperqa.clients.semantic_scholar, "SEMANTIC_SCHOLAR_API_REQUEST_TIMEOUT", 0.05
    )

    async with aiohttp.ClientSession() as session:
        client = DocMetadataClient(session)
        details = await client.query(
            doi="10.1007/s40278-023-41815-2",
            fields=["doi", "title"],
        )
    assert details is None, "Should return None for timeout"


@pytest.mark.asyncio()
async def test_bad_init():
    with pytest.raises(
        ValueError, match="At least one MetadataProvider must be provided."
    ):
        client = DocMetadataClient(clients=[])  # noqa: F841


@pytest.mark.vcr()
@pytest.mark.asyncio()
async def test_ensure_sequential_run(caplog):
    caplog.set_level(logging.DEBUG)
    # were using a DOI that is NOT in crossref, but running the crossref client first
    # we will ensure that both are run sequentially

    async with aiohttp.ClientSession() as session:
        client = DocMetadataClient(
            session=session,
            clients=cast(
                Sequence[
                    Collection[
                        type[MetadataPostProcessor[Any]] | type[MetadataProvider[Any]]
                    ]
                ],
                [[CrossrefProvider], [SemanticScholarProvider]],
            ),
        )
        details = await client.query(
            doi="10.48550/arxiv.2312.07559",
            fields=["doi", "title"],
        )
        assert details, "Should find the right DOI in the second client"
        record_indices = {"crossref": -1, "semantic_scholar": -1}
        for n, record in enumerate(caplog.records):
            if "CrossrefProvider" in record.msg:
                record_indices["crossref"] = n
            if "SemanticScholarProvider" in record.msg:
                record_indices["semantic_scholar"] = n
        assert (
            record_indices["crossref"] < record_indices["semantic_scholar"]
        ), "Crossref should run first"


@pytest.mark.vcr()
@pytest.mark.asyncio()
async def test_ensure_sequential_run_early_stop(caplog):
    caplog.set_level(logging.DEBUG)
    # now we should stop after hitting s2
    async with aiohttp.ClientSession() as session:
        client = DocMetadataClient(
            session=session,
            clients=cast(
                Sequence[
                    Collection[
                        type[MetadataPostProcessor[Any]] | type[MetadataProvider[Any]]
                    ]
                ],
                [[SemanticScholarProvider], [CrossrefProvider]],
            ),
        )
        details = await client.query(
            doi="10.48550/arxiv.2312.07559",
            fields=["doi", "title"],
        )
        assert details, "Should find the right DOI in the second client"
        record_indices = {"crossref": -1, "semantic_scholar": -1, "early_stop": -1}
        for n, record in enumerate(caplog.records):
            if "CrossrefProvider" in record.msg:
                record_indices["crossref"] = n
            if "SemanticScholarProvider" in record.msg:
                record_indices["semantic_scholar"] = n
            if "stopping early." in record.msg:
                record_indices["early_stop"] = n
        assert (
            record_indices["crossref"] == -1
        ), "Crossref should be index -1 i.e. not found"
        assert (
            record_indices["semantic_scholar"] != -1
        ), "Semantic Scholar should be found"
        assert record_indices["early_stop"] != -1, "We should stop early."



from __future__ import annotations

import logging
import os
from typing import Any, cast
from unittest.mock import patch

from langchain.agents import AgentExecutor, BaseSingleActionAgent, ZeroShotAgent
from langchain.agents.openai_functions_agent.base import OpenAIFunctionsAgent
from langchain_community.callbacks import OpenAICallbackHandler
from langchain_core.agents import AgentAction, AgentFinish
from langchain_core.callbacks import BaseCallbackHandler, BaseCallbackManager, Callbacks
from langchain_core.messages import SystemMessage
from langchain_openai import ChatOpenAI
from rich.console import Console

from ..docs import Docs
from ..types import Answer
from ..utils import pqa_directory
from .helpers import openai_get_search_query, table_formatter, update_doc_models
from .models import (
    AgentCallback,
    AgentStatus,
    AnswerResponse,
    QueryRequest,
    SimpleProfiler,
)
from .search import SearchDocumentStorage, SearchIndex
from .tools import (
    EmptyDocsError,
    GatherEvidenceTool,
    GenerateAnswerTool,
    PaperSearchTool,
    SharedToolState,
    query_to_tools,
    status,
)

logger = logging.getLogger(__name__)
agent_logger = logging.getLogger(__name__ + ".agent_callers")


async def agent_query(
    query: str | QueryRequest,
    docs: Docs | None = None,
    agent_type: str = "OpenAIFunctionsAgent",
    verbosity: int = 0,
    index_directory: str | os.PathLike | None = None,
) -> AnswerResponse:

    if isinstance(query, str):
        query = QueryRequest(query=query)

    if docs is None:
        docs = Docs()

    if index_directory is None:
        index_directory = pqa_directory("indexes")

    # in-place modification of the docs object to match query
    update_doc_models(
        docs,
        query,
    )

    search_index = SearchIndex(
        fields=SearchIndex.REQUIRED_FIELDS | {"question"},
        index_name="answers",
        index_directory=index_directory,
        storage=SearchDocumentStorage.JSON_MODEL_DUMP,
    )

    response = await run_agent(docs, query, agent_type)

    agent_logger.debug(f"agent_response: {response}")
    truncation_chars = 1_000_000 if verbosity > 1 else 1500 * (verbosity + 1)
    agent_logger.info(
        f"[bold blue]Answer: {response.answer.answer[:truncation_chars]}"
        f'{"...(truncated)" if len(response.answer.answer) > truncation_chars else ""}[/bold blue]'
    )

    await search_index.add_document(
        {
            "file_location": str(response.answer.id),
            "body": response.answer.answer or "",
            "question": response.answer.question,
        },
        document=response,
    )

    await search_index.save_index()

    return response


async def run_agent(
    docs: Docs,
    query: QueryRequest,
    agent_type: str = "OpenAIFunctionsAgent",
) -> AnswerResponse:
    """
    Run an agent.

    Args:
        docs: Docs to run upon.
        query: Query to answer.
        websocket: Websocket to send JSON data and receive text.
        agent_type: Agent type to pass to AgentType.get_agent, or "fake" to TODOC.

    Returns:
        Tuple of resultant answer, token counts, and agent status.
    """
    profiler = SimpleProfiler()
    outer_profile_name = f"agent-{agent_type}-{query.agent_llm}"
    profiler.start(outer_profile_name)

    logger.info(
        f"Beginning agent {agent_type!r} run with question {query.query!r} and full"
        f" query {query.model_dump()}."
    )

    if agent_type == "fake":
        answer, agent_status = await run_fake_agent(query, docs)
    else:
        answer, agent_status = await run_langchain_agent(
            query, docs, agent_type, profiler
        )

    if "cannot answer" in answer.answer.lower() and agent_status != AgentStatus.TIMEOUT:
        agent_status = AgentStatus.UNSURE
    # stop after, so overall isn't reported as long-running step.
    logger.info(
        f"Finished agent {agent_type!r} run with question {query.query!r} and status"
        f" {agent_status}."
    )
    return AnswerResponse(
        answer=answer,
        usage=answer.token_counts,
        status=agent_status,
    )


async def run_fake_agent(
    query: QueryRequest,
    docs: Docs,
) -> tuple[Answer, AgentStatus]:
    answer = Answer(question=query.query, dockey_filter=set(), id=query.id)
    tools = query_to_tools(query, state=SharedToolState(docs=docs, answer=answer))
    search_tool = cast(
        PaperSearchTool,
        next(
            filter(
                lambda x: x.name == PaperSearchTool.__fields__["name"].default, tools
            )
        ),
    )
    gather_evidence_tool = cast(
        GatherEvidenceTool,
        next(
            filter(
                lambda x: x.name == GatherEvidenceTool.__fields__["name"].default, tools
            )
        ),
    )

    generate_answer_tool = cast(
        GenerateAnswerTool,
        next(
            filter(
                lambda x: x.name == GenerateAnswerTool.__fields__["name"].default, tools
            )
        ),
    )
    # seed docs with keyword search
    for search in await openai_get_search_query(
        answer.question, llm=query.llm, count=3
    ):
        await search_tool.arun(search)

    await gather_evidence_tool.arun(tool_input=answer.question)

    await generate_answer_tool.arun(tool_input=answer.question)

    return answer, AgentStatus.SUCCESS


LANGCHAIN_AGENT_TYPES: dict[str, type[BaseSingleActionAgent]] = {
    "ReactAgent": ZeroShotAgent,
    "OpenAIFunctionsAgent": OpenAIFunctionsAgent,
}


async def run_langchain_agent(
    query: QueryRequest,
    docs: Docs,
    agent_type: str,
    profiler: SimpleProfiler,
    timeout: float | None = None,  # noqa: ASYNC109
) -> tuple[Answer, AgentStatus]:
    answer = Answer(question=query.query, dockey_filter=set(), id=query.id)
    shared_callbacks: list[BaseCallbackHandler] = [
        AgentCallback(
            profiler, name=f"step-{agent_type}-{query.agent_llm}", answer_id=answer.id
        ),
    ]
    tools = query_to_tools(
        query,
        state=SharedToolState(docs=docs, answer=answer),
        callbacks=shared_callbacks,
    )
    try:
        search_tool = next(
            filter(
                lambda x: x.name == PaperSearchTool.__fields__["name"].default, tools
            )
        )
    except StopIteration:
        search_tool = None
    answer_tool = cast(
        GenerateAnswerTool,
        next(
            filter(
                lambda x: x.name == GenerateAnswerTool.__fields__["name"].default, tools
            )
        ),
    )

    # optionally use the search tool before the agent
    if search_tool is not None and query.agent_tools.should_pre_search:
        logger.debug("Running search tool before agent choice.")
        await search_tool.arun(answer.question)
    else:
        logger.debug("Skipping search tool before agent choice.")

    llm = ChatOpenAI(
        model=query.agent_llm,
        request_timeout=timeout or query.agent_tools.timeout / 2.0,
        temperature=query.temperature,
    )
    agent_status = AgentStatus.SUCCESS
    cost_callback = OpenAICallbackHandler()
    agent_instance = LANGCHAIN_AGENT_TYPES[agent_type].from_llm_and_tools(
        llm,
        tools,
        system_message=(
            SystemMessage(content=query.agent_tools.agent_system_prompt)
            if query.agent_tools.agent_system_prompt
            else None
        ),
    )
    orig_aplan = agent_instance.aplan
    agent_exec_instance = AgentExecutor.from_agent_and_tools(
        tools=tools,
        agent=agent_instance,
        return_intermediate_steps=True,
        handle_parsing_errors=True,
        max_execution_time=query.agent_tools.timeout,
        callbacks=[*shared_callbacks, cost_callback],
        **(query.agent_tools.agent_config or {}),
    )

    async def aplan_with_injected_callbacks(
        intermediate_steps: list[tuple[AgentAction, str]],
        callbacks: Callbacks = None,
        **kwargs,
    ) -> AgentAction | AgentFinish:
        # Work around https://github.com/langchain-ai/langchain/issues/22703
        for callback in cast(list[BaseCallbackHandler], agent_exec_instance.callbacks):
            cast(BaseCallbackManager, callbacks).add_handler(callback, inherit=False)
        return await orig_aplan(intermediate_steps, callbacks, **kwargs)

    try:
        # Patch at instance (not class) level to avoid concurrency issues, and we have
        # to patch the dict to work around Pydantic's BaseModel.__setattr__'s validations
        with patch.dict(
            agent_instance.__dict__, {"aplan": aplan_with_injected_callbacks}
        ):
            call_response = await agent_exec_instance.ainvoke(
                input={
                    # NOTE: str.format still works even if the prompt doesn't have
                    # template fields like 'status' or 'gen_answer_tool_name'
                    "input": query.agent_tools.agent_prompt.format(
                        question=answer.question,
                        status=await status(docs, answer),
                        gen_answer_tool_name=answer_tool.name,
                    )
                }
            )
    except TimeoutError:
        call_response = {"output": "Agent stopped", "intermediate_steps": []}
    except EmptyDocsError:
        call_response = {
            "output": "Agent failed due to failed search",
            "intermediate_steps": [],
        }
        agent_status = AgentStatus.FAIL

    async with profiler.timer("agent-accounting"):
        # TODO: move agent trace to LangChain callback
        if "Agent stopped" in call_response["output"]:
            # Log that this agent has gone over timeout, and then answer directly
            logger.warning(
                f"Agent timeout after {query.agent_tools.timeout}-sec, just answering."
            )
            await answer_tool.arun(answer.question)
            agent_status = AgentStatus.TIMEOUT

    return answer, agent_status


async def search(
    query: str,
    index_name: str = "answers",
    index_directory: str | os.PathLike | None = None,
) -> list[tuple[AnswerResponse, str] | tuple[Any, str]]:

    search_index = SearchIndex(
        ["file_location", "body", "question"],
        index_name=index_name,
        index_directory=index_directory or pqa_directory("indexes"),
        storage=SearchDocumentStorage.JSON_MODEL_DUMP,
    )

    results = [
        (AnswerResponse(**a[0]) if index_name == "answers" else a[0], a[1])
        for a in await search_index.query(query=query, keep_filenames=True)
    ]

    if results:
        console = Console(record=True)
        # Render the table to a string
        console.print(table_formatter(results))
    else:
        agent_logger.info("No results found.")

    return results



from __future__ import annotations

import logging
import os
from datetime import datetime
from enum import IntEnum, auto
from http import HTTPStatus
from itertools import starmap
from typing import Any, Collection

import aiohttp

from ..types import DocDetails
from ..utils import (
    _get_with_retrying,
    clean_upbibtex,
    strings_similarity,
    union_collections_to_ordered_list,
)
from .client_models import DOIOrTitleBasedProvider, DOIQuery, TitleAuthorQuery
from .crossref import doi_to_bibtex
from .exceptions import DOINotFoundError

logger = logging.getLogger(__name__)

# map from S2 fields to those in the DocDetails model
# allows users to specify which fields to include in the response
SEMANTIC_SCHOLAR_API_MAPPING: dict[str, Collection[str]] = {
    "title": {"title"},
    "doi": {"externalIds"},
    "authors": {"authors"},
    "publication_date": {"publicationDate"},
    "year": {"year"},
    "volume": {"journal"},
    "pages": {"journal"},
    "journal": {"journal"},
    "url": {"url", "openAccessPdf"},
    "bibtex": {"citationStyles"},
    "doi_url": {"url"},
    "other": {"isOpenAccess", "influentialCitationCount", "publicationTypes", "venue"},
    "citation_count": {"citationCount"},
    "source_quality": {"journal"},
}
SEMANTIC_SCHOLAR_API_REQUEST_TIMEOUT = 10.0
SEMANTIC_SCHOLAR_API_FIELDS: str = ",".join(
    union_collections_to_ordered_list(SEMANTIC_SCHOLAR_API_MAPPING.values())
)
SEMANTIC_SCHOLAR_BASE_URL = "https://api.semanticscholar.org"
SEMANTIC_SCHOLAR_HEADER_KEY = "x-api-key"


class SematicScholarSearchType(IntEnum):
    DEFAULT = auto()
    PAPER = auto()
    PAPER_RECOMMENDATIONS = auto()
    DOI = auto()
    FUTURE_CITATIONS = auto()
    PAST_REFERENCES = auto()
    GOOGLE = auto()
    MATCH = auto()

    def make_url_params(  # noqa: PLR0911
        self,
        params: dict[str, Any],
        query: str = "",
        offset: int = 0,
        limit: int = 1,
        include_base_url: bool = True,
    ) -> tuple[str, dict[str, Any]]:
        """
        Make the target URL and in-place update the input URL parameters.

        Args:
            params: URL parameters to in-place update.
            query: Either a search query or a Semantic Scholar paper ID.
            offset: Offset to place in the URL parameters for the default search type.
            limit: Limit to place in the URL parameters for some search types.
            include_base_url: Set True (default) to include the base URL.

        Returns:
            Two-tuple of URL and URL parameters.
        """
        base = SEMANTIC_SCHOLAR_BASE_URL if include_base_url else ""
        if self == SematicScholarSearchType.DEFAULT:
            params["query"] = query.replace("-", " ")
            params["offset"] = offset
            params["limit"] = limit
            return f"{base}/graph/v1/paper/search", params
        if self == SematicScholarSearchType.PAPER:
            return f"{base}/graph/v1/paper/{query}", params
        if self == SematicScholarSearchType.PAPER_RECOMMENDATIONS:
            return f"{base}/recommendations/v1/papers/forpaper/{query}", params
        if self == SematicScholarSearchType.DOI:
            return f"{base}/graph/v1/paper/DOI:{query}", params
        if self == SematicScholarSearchType.FUTURE_CITATIONS:
            params["limit"] = limit
            return f"{base}/graph/v1/paper/{query}/citations", params
        if self == SematicScholarSearchType.PAST_REFERENCES:
            params["limit"] = limit
            return f"{base}/graph/v1/paper/{query}/references", params
        if self == SematicScholarSearchType.GOOGLE:
            params["limit"] = 1
            return f"{base}/graph/v1/paper/search", params
        if self == SematicScholarSearchType.MATCH:
            return f"{base}/graph/v1/paper/search/match", params
        raise NotImplementedError


def s2_authors_match(authors: list[str], data: dict):
    """Check if the authors in the data match the authors in the paper."""
    AUTHOR_NAME_MIN_LENGTH = 2
    s2_authors_noinit = [
        " ".join([w for w in a["name"].split() if len(w) > AUTHOR_NAME_MIN_LENGTH])
        for a in data["authors"]
    ]
    authors_noinit = [
        " ".join([w for w in a.split() if len(w) > AUTHOR_NAME_MIN_LENGTH])
        for a in authors
    ]
    # Note: we expect the number of authors to be possibly different
    return any(
        starmap(
            lambda x, y: x in y or y in x,
            zip(s2_authors_noinit, authors_noinit, strict=False),
        )
    )


async def parse_s2_to_doc_details(
    paper_data: dict, session: aiohttp.ClientSession
) -> DocDetails:

    bibtex_source = "self_generated"

    if "data" in paper_data:
        paper_data = paper_data["data"][0]

    if "ArXiv" in paper_data["externalIds"]:
        doi = "10.48550/arXiv." + paper_data["externalIds"]["ArXiv"]
    elif "DOI" in paper_data["externalIds"]:
        doi = paper_data["externalIds"]["DOI"]
    else:
        raise DOINotFoundError(f"Could not find DOI for {paper_data}.")

    # Should we give preference to auto-generation?
    if not (
        bibtex := clean_upbibtex(paper_data.get("citationStyles", {}).get("bibtex", ""))
    ):
        try:
            bibtex = await doi_to_bibtex(doi, session)
            bibtex_source = "crossref"
        except DOINotFoundError:
            bibtex = None
    else:
        bibtex_source = "semantic_scholar"

    publication_date = None
    if paper_data.get("publicationDate"):
        publication_date = datetime.strptime(paper_data["publicationDate"], "%Y-%m-%d")

    doc_details = DocDetails(  # type: ignore[call-arg]
        key=None if not bibtex else bibtex.split("{")[1].split(",")[0],
        bibtex_type="article",  # s2 should be basically all articles
        bibtex=bibtex,
        authors=[author["name"] for author in paper_data.get("authors", [])],
        publication_date=publication_date,
        year=paper_data.get("year"),
        volume=paper_data.get("journal", {}).get("volume"),
        pages=paper_data.get("journal", {}).get("pages"),
        journal=paper_data.get("journal", {}).get("name"),
        url=(paper_data.get("openAccessPdf") or {}).get("url"),
        title=paper_data.get("title"),
        citation_count=paper_data.get("citationCount"),
        doi=paper_data.get("externalIds", {}).get("DOI"),
        other={},  # Initialize empty dict for other fields
    )

    # Add any additional fields to the 'other' dict
    for key, value in (
        paper_data
        | {"client_source": ["semantic_scholar"], "bibtex_source": [bibtex_source]}
    ).items():
        if key not in doc_details.model_fields:
            doc_details.other[key] = value

    return doc_details


def semantic_scholar_headers() -> dict[str, str]:
    """Semantic Scholar API key if available, otherwise nothing."""
    if api_key := os.environ.get("SEMANTIC_SCHOLAR_API_KEY"):
        return {SEMANTIC_SCHOLAR_HEADER_KEY: api_key}
    logger.warning(
        "SEMANTIC_SCHOLAR_API_KEY environment variable not set. Semantic Scholar API rate limits may apply."
    )
    return {}


async def s2_title_search(
    title: str,
    session: aiohttp.ClientSession,
    authors: list[str] | None = None,
    title_similarity_threshold: float = 0.75,
    fields: str = SEMANTIC_SCHOLAR_API_FIELDS,
) -> DocDetails:
    """Reconcile DOI from Semantic Scholar - which only checks title. So we manually check authors."""
    if authors is None:
        authors = []
    endpoint, params = SematicScholarSearchType.MATCH.make_url_params(
        params={"query": title, "fields": fields}
    )

    data = await _get_with_retrying(
        url=endpoint,
        params=params,
        session=session,
        headers=semantic_scholar_headers(),
        timeout=SEMANTIC_SCHOLAR_API_REQUEST_TIMEOUT,
        http_exception_mappings={
            HTTPStatus.NOT_FOUND: DOINotFoundError(f"Could not find DOI for {title}.")
        },
    )

    if authors and not s2_authors_match(authors, data["data"][0]):
        raise DOINotFoundError(f"Could not find DOI for {title} - author disagreement.")
    # need to check if nested under a 'data' key or not (depends on filtering)
    if (
        strings_similarity(
            data.get("title") if "data" not in data else data["data"][0]["title"],
            title,
        )
        < title_similarity_threshold
    ):
        raise DOINotFoundError(
            f"Semantic scholar results did not match for title {title!r}."
        )
    return await parse_s2_to_doc_details(data, session)


async def get_s2_doc_details_from_doi(
    doi: str | None,
    session: aiohttp.ClientSession,
    fields: Collection[str] | None = None,
) -> DocDetails:
    """Get paper details from Semantic Scholar given a DOI."""
    # should always be string, runtime error catch
    if doi is None:
        raise ValueError("Valid DOI must be provided.")

    if fields:
        s2_fields = ",".join(
            union_collections_to_ordered_list(
                SEMANTIC_SCHOLAR_API_MAPPING[f]
                for f in fields
                if f in SEMANTIC_SCHOLAR_API_MAPPING
            )
        )
    else:
        s2_fields = SEMANTIC_SCHOLAR_API_FIELDS

    details = await _get_with_retrying(
        url=f"{SEMANTIC_SCHOLAR_BASE_URL}/graph/v1/paper/DOI:{doi}",
        params={"fields": s2_fields},
        session=session,
        headers=semantic_scholar_headers(),
        timeout=SEMANTIC_SCHOLAR_API_REQUEST_TIMEOUT,
        http_exception_mappings={
            HTTPStatus.NOT_FOUND: DOINotFoundError(f"Could not find DOI for {doi}.")
        },
    )

    return await parse_s2_to_doc_details(details, session)


async def get_s2_doc_details_from_title(
    title: str | None,
    session: aiohttp.ClientSession,
    authors: list[str] | None = None,
    fields: Collection[str] | None = None,
    title_similarity_threshold: float = 0.75,
) -> DocDetails:
    """Get paper details from Semantic Scholar given a title.

    Optionally match against authors if provided.
    """
    if title is None:
        raise ValueError("Valid title must be provided.")
    if authors is None:
        authors = []
    if fields:
        s2_fields = ",".join(
            union_collections_to_ordered_list(
                SEMANTIC_SCHOLAR_API_MAPPING[f]
                for f in fields
                if f in SEMANTIC_SCHOLAR_API_MAPPING
            )
        )
    else:
        s2_fields = SEMANTIC_SCHOLAR_API_FIELDS

    return await s2_title_search(
        title,
        authors=authors,
        session=session,
        title_similarity_threshold=title_similarity_threshold,
        fields=s2_fields,
    )


class SemanticScholarProvider(DOIOrTitleBasedProvider):
    async def _query(self, query: TitleAuthorQuery | DOIQuery) -> DocDetails | None:
        if isinstance(query, DOIQuery):
            return await get_s2_doc_details_from_doi(
                doi=query.doi, session=query.session, fields=query.fields
            )
        return await get_s2_doc_details_from_title(
            title=query.title,
            authors=query.authors,
            session=query.session,
            title_similarity_threshold=query.title_similarity_threshold,
            fields=query.fields,
        )



from pathlib import Path

import pytest

try:
    from paperqa.agents import ask, build_index, clear, search_query, set_setting, show
    from paperqa.agents.models import AnswerResponse
    from paperqa.agents.search import SearchIndex
except ImportError:
    pytest.skip("agents module is not installed", allow_module_level=True)


def test_cli_set(agent_index_dir: Path):  # noqa: ARG001
    set_setting("temperature", "0.5")
    assert show("temperature") == "0.5", "Temperature not properly set"

    with pytest.raises(ValueError) as excinfo:  # noqa: PT011
        set_setting("temperature", "abc")
    assert "temperature (with value abc) is not a valid setting." in str(excinfo.value)

    # ensure we can do nested settings
    set_setting("agent_tools.paper_directory", "my_directory")
    assert (
        show("agent_tools.paper_directory") == "my_directory"
    ), "Nested setting not properly set"

    # ensure we can set settings which fail specific validations
    # normally we'd get a failure for model mixing, but this is reserved for runtime
    set_setting("llm", "claude-3-5-sonnet-20240620")
    assert show("llm") == "claude-3-5-sonnet-20240620", "Setting not properly set"

    # test that we're able to collection structures like lists
    set_setting(
        "parsing_configuration.ordered_parser_preferences", '["paperqa_default"]'
    )
    assert show("parsing_configuration.ordered_parser_preferences") == [
        "paperqa_default"
    ], "List setting not properly set"


@pytest.mark.asyncio
async def test_cli_show(agent_index_dir: Path):

    # make empty index
    assert not show("indexes"), "No indexes should be present"

    # creates a new index/file
    si = SearchIndex(index_directory=agent_index_dir)
    await si.init_directory()

    set_setting("temperature", "0.5")
    set_setting("agent_tools.paper_directory", "my_directory")

    assert show("temperature") == "0.5", "Temperature not properly set"

    assert not show("fake_variable"), "Fake variable should not be set"

    assert show("all") == {
        "temperature": "0.5",
        "agent_tools": {"paper_directory": "my_directory"},
    }, "All settings not properly set"

    assert show("indexes") == ["pqa_index"], "Index not properly set"

    assert show("answers") == [], "Answers should be empty"


@pytest.mark.asyncio
async def test_cli_clear(agent_index_dir: Path):

    set_setting("temperature", "0.5")
    assert show("temperature") == "0.5", "Temperature not properly set"

    clear("temperature")
    assert show("temperature") is None, "Temperature not properly cleared"

    # set a nested variable
    set_setting("prompts.qa", "Answer my question!")
    assert show("prompts.qa") == "Answer my question!", "Prompt not properly set"
    clear("prompts.qa")
    assert show("prompts.qa") is None, "Prompt not properly cleared"

    # creates a new index/file
    si = SearchIndex(index_directory=agent_index_dir)
    await si.init_directory()

    clear("pqa_index", index=True)

    assert show("indexes") == [], "Index not properly cleared"


def test_cli_ask(agent_index_dir: Path):
    set_setting("consider_sources", "10")
    set_setting("max_sources", "2")
    set_setting("embedding", "sparse")
    set_setting("agent_tools.search_count", "1")
    answer = ask(
        "How can you use XAI for chemical property prediction?",
        directory=Path(__file__).parent,
        index_directory=agent_index_dir,
    )
    assert isinstance(answer, AnswerResponse), "Answer not properly returned"
    assert (
        "I cannot answer" not in answer.answer.answer
    ), "An answer should be generated."
    assert len(answer.answer.context) >= 1, "No contexts were found."
    answers = search_query(
        "How can you use XAI for chemical property prediction?",
        index_directory=agent_index_dir,
    )
    answer = answers[0][0]
    assert isinstance(answer, AnswerResponse), "Answer not properly returned"
    assert (
        "I cannot answer" not in answer.answer.answer
    ), "An answer should be generated."
    assert len(answer.answer.context) >= 1, "No contexts were found."
    assert len(show("answers")) == 1, "An answer should be returned"


def test_cli_index(agent_index_dir: Path, caplog):

    build_index(directory=Path(__file__).parent, index_directory=agent_index_dir)

    caplog.clear()

    ask(
        "How can you use XAI for chemical property prediction?",
        directory=Path(__file__).parent,
        index_directory=agent_index_dir,
    )

    # ensure we have no indexing logs after starting the search
    for record in caplog.records:
        if "Metadata not found" in record.msg:
            raise AssertionError(
                "Indexing logs should not be present after search starts"
            )

    caplog.clear()

    # running again should not trigger any indexing
    build_index(directory=Path(__file__).parent, index_directory=agent_index_dir)
    assert not caplog.records, "Indexing should not be triggered again"

    # now we want to change the settings
    set_setting("embedding", "sparse")

    # running again should now re-trigger an indexing
    build_index(directory=Path(__file__).parent, index_directory=agent_index_dir)

    assert caplog.records, "Indexing should be triggered again"



from __future__ import annotations

import copy
import logging
from typing import Any, Collection, Coroutine, Sequence

import aiohttp
from pydantic import BaseModel, ConfigDict

from ..types import Doc, DocDetails
from ..utils import gather_with_concurrency
from .client_models import MetadataPostProcessor, MetadataProvider
from .crossref import CrossrefProvider
from .journal_quality import JournalQualityPostProcessor
from .semantic_scholar import SemanticScholarProvider

logger = logging.getLogger(__name__)

ALL_CLIENTS: (
    Collection[type[MetadataPostProcessor | MetadataProvider]]
    | Sequence[Collection[type[MetadataPostProcessor | MetadataProvider]]]
) = {
    CrossrefProvider,
    SemanticScholarProvider,
    JournalQualityPostProcessor,
}


class DocMetadataTask(BaseModel):
    """Holder for provider and processor tasks."""

    providers: Collection[MetadataProvider]
    processors: Collection[MetadataPostProcessor]

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def provider_queries(
        self, query: dict
    ) -> list[Coroutine[Any, Any, DocDetails | None]]:
        return [p.query(query) for p in self.providers]

    def processor_queries(
        self, doc_details: DocDetails, session: aiohttp.ClientSession
    ) -> list[Coroutine[Any, Any, DocDetails]]:
        return [
            p.process(copy.copy(doc_details), session=session) for p in self.processors
        ]

    def __repr__(self) -> str:
        return (
            f"DocMetadataTask(providers={self.providers}, processors={self.processors})"
        )


class DocMetadataClient:
    def __init__(
        self,
        session: aiohttp.ClientSession | None = None,
        clients: (
            Collection[type[MetadataPostProcessor | MetadataProvider]]
            | Sequence[Collection[type[MetadataPostProcessor | MetadataProvider]]]
        ) = ALL_CLIENTS,
    ) -> None:
        """Metadata client for querying multiple metadata providers and processors.

        Args:
            session: outer scope aiohttp session to allow for connection pooling
            clients: list of MetadataProvider and MetadataPostProcessor classes to query;
                if nested, will query in order looking for termination criteria after each.
                Will terminate early if either DocDetails.is_hydration_needed is False OR if
                all requested fields are present in the DocDetails object.

        """
        self._session = session
        self.tasks: list[DocMetadataTask] = []

        # first see if we are nested; i.e. we want order
        if isinstance(clients, Sequence) and all(
            isinstance(sub_clients, Collection) for sub_clients in clients
        ):
            for sub_clients in clients:
                self.tasks.append(
                    DocMetadataTask(
                        providers=[
                            c() for c in sub_clients if issubclass(c, MetadataProvider)
                        ],
                        processors=[
                            c()
                            for c in sub_clients
                            if issubclass(c, MetadataPostProcessor)
                        ],
                    )
                )
        # otherwise, we are a flat collection
        if not self.tasks and all(not isinstance(c, Collection) for c in clients):
            self.tasks.append(
                DocMetadataTask(
                    providers=[c() for c in clients if issubclass(c, MetadataProvider)],  # type: ignore[operator, arg-type]
                    processors=[
                        c() for c in clients if issubclass(c, MetadataPostProcessor)  # type: ignore[operator, arg-type]
                    ],
                )
            )

        if not self.tasks or (self.tasks and not self.tasks[0].providers):
            raise ValueError("At least one MetadataProvider must be provided.")

    async def query(self, **kwargs) -> DocDetails | None:

        session = aiohttp.ClientSession() if self._session is None else self._session

        query_args = kwargs if "session" in kwargs else kwargs | {"session": session}

        doc_details: DocDetails | None = None

        for ti, task in enumerate(self.tasks):
            logger.debug(
                f"Attempting to populate metadata query: {query_args} via {task}"
            )

            # first query all client_models and aggregate the results
            doc_details = (
                sum(
                    p
                    for p in (
                        await gather_with_concurrency(
                            len(task.providers), task.provider_queries(query_args)
                        )
                    )
                    if p
                )
                or None
            )

            # then process and re-aggregate the results
            if doc_details and task.processors:
                doc_details = sum(
                    await gather_with_concurrency(
                        len(task.processors),
                        task.processor_queries(doc_details, session),
                    )
                )

            if doc_details and not doc_details.is_hydration_needed(
                inclusion=kwargs.get("fields", [])
            ):
                logger.debug(
                    "All requested fields are present in the DocDetails "
                    f"object{', stopping early.' if ti != len(self.tasks) - 1 else '.'}"
                )
                break

        if self._session is None:
            await session.close()

        return doc_details

    async def bulk_query(
        self, queries: Collection[dict[str, Any]], concurrency: int = 10
    ) -> list[DocDetails]:
        return await gather_with_concurrency(
            concurrency, [self.query(**kwargs) for kwargs in queries]
        )

    async def upgrade_doc_to_doc_details(self, doc: Doc, **kwargs) -> DocDetails:

        # note we have some extra fields which may have come from reading the doc text,
        # but aren't in the doc object, we add them here too.
        extra_fields = {
            k: v for k, v in kwargs.items() if k in {"title", "authors", "doi"}
        }
        # abuse our doc_details object to be an int if it's empty
        # our __add__ operation supports int by doing nothing
        extra_doc: int | DocDetails = (
            0 if not extra_fields else DocDetails(**extra_fields)
        )

        if doc_details := await self.query(**kwargs):
            if doc.overwrite_fields_from_metadata:
                return extra_doc + doc_details

            # hard overwrite the details from the prior object
            doc_details.dockey = doc.dockey
            doc_details.doc_id = doc.dockey
            doc_details.docname = doc.docname
            doc_details.key = doc.docname
            doc_details.citation = doc.citation
            return extra_doc + doc_details

        # if we can't get metadata, just return the doc, but don't overwrite any fields
        prior_doc = doc.model_dump()
        prior_doc["overwrite_fields_from_metadata"] = False
        return DocDetails(**(prior_doc | extra_fields))



from __future__ import annotations

import csv
import logging
import os
from typing import Any

from pydantic import ValidationError

from ..types import DocDetails
from .client_models import JournalQuery, MetadataPostProcessor

logger = logging.getLogger(__name__)


# TODO: refresh script for journal quality data


class JournalQualityPostProcessor(MetadataPostProcessor[JournalQuery]):
    def __init__(self, journal_quality_path: os.PathLike | str | None = None) -> None:
        if journal_quality_path is None:
            # Construct the path relative to module
            self.journal_quality_path = str(
                os.path.join(
                    os.path.dirname(__file__), "client_data", "journal_quality.csv"
                )
            )
        else:
            self.journal_quality_path = str(journal_quality_path)
        self.data: dict[str, Any] | None = None

    def load_data(self) -> None:
        self.data = {}
        with open(self.journal_quality_path, newline="", encoding="utf-8") as csvfile:
            for row in csv.DictReader(csvfile):
                self.data[row["clean_name"]] = int(row["quality"])

    async def _process(
        self, query: JournalQuery, doc_details: DocDetails
    ) -> DocDetails:
        if not self.data:
            self.load_data()
        # docname can be blank since the validation will add it
        # remember, if both have docnames (i.e. key) they are
        # wiped and re-generated with resultant data
        return doc_details + DocDetails(  # type: ignore[call-arg]
            source_quality=self.data.get(query.journal.casefold(), -1)  # type: ignore[union-attr]
        )

    def query_creator(self, doc_details: DocDetails, **kwargs) -> JournalQuery | None:
        try:
            return JournalQuery(journal=doc_details.journal, **kwargs)
        except ValidationError:
            logger.debug(
                "Must have a valid journal name to query journal quality data."
            )
            return None



from __future__ import annotations

import csv
import json
import logging
import os
import pathlib
import pickle
import zlib
from enum import Enum, auto
from io import StringIO
from typing import Any, ClassVar, Collection
from uuid import UUID

import anyio
from pydantic import BaseModel
from tantivy import Document, Index, Schema, SchemaBuilder, Searcher
from tenacity import (
    RetryError,
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_random_exponential,
)

from ..docs import Docs
from ..types import DocDetails
from ..utils import hexdigest, pqa_directory
from .models import SupportsPickle

logger = logging.getLogger(__name__)

PQA_INDEX_ABSOLUTE_PATHS = (
    os.environ.get("PQA_INDEX_ABSOLUTE_PATHS", "true").lower() == "true"
)


class AsyncRetryError(Exception):
    """Flags a retry for another tenacity attempt."""


class RobustEncoder(json.JSONEncoder):
    """JSON encoder that can handle UUID and set objects."""

    def default(self, obj):
        if isinstance(obj, UUID):
            # if the obj is uuid, we simply return the value of uuid
            return str(obj)
        if isinstance(obj, set):
            return list(obj)
        return json.JSONEncoder.default(self, obj)


class SearchDocumentStorage(str, Enum):
    JSON_MODEL_DUMP = auto()
    PICKLE_COMPRESSED = auto()
    PICKLE_UNCOMPRESSED = auto()

    def extension(self) -> str:
        if self == SearchDocumentStorage.JSON_MODEL_DUMP:
            return "json"
        if self == SearchDocumentStorage.PICKLE_COMPRESSED:
            return "zip"
        return "pkl"

    def write_to_string(self, data: BaseModel | SupportsPickle) -> bytes:
        if self == SearchDocumentStorage.JSON_MODEL_DUMP:
            if isinstance(data, BaseModel):
                return json.dumps(data.model_dump(), cls=RobustEncoder).encode("utf-8")
            raise ValueError("JSON_MODEL_DUMP requires a BaseModel object.")
        if self == SearchDocumentStorage.PICKLE_COMPRESSED:
            return zlib.compress(pickle.dumps(data))
        return pickle.dumps(data)

    def read_from_string(self, data: str | bytes) -> BaseModel | SupportsPickle:
        if self == SearchDocumentStorage.JSON_MODEL_DUMP:
            return json.loads(data)
        if self == SearchDocumentStorage.PICKLE_COMPRESSED:
            return pickle.loads(zlib.decompress(data))  # type: ignore[arg-type] # noqa: S301
        return pickle.loads(data)  # type: ignore[arg-type] # noqa: S301


class SearchIndex:

    REQUIRED_FIELDS: ClassVar[set[str]] = {"file_location", "body"}

    def __init__(
        self,
        fields: Collection[str] | None = None,
        index_name: str = "pqa_index",
        index_directory: str | os.PathLike | None = None,
        storage: SearchDocumentStorage = SearchDocumentStorage.PICKLE_COMPRESSED,
    ):
        if fields is None:
            fields = self.REQUIRED_FIELDS
        self.fields = fields
        if not all(f in self.fields for f in self.REQUIRED_FIELDS):
            raise ValueError(
                f"{self.REQUIRED_FIELDS} must be included in search index fields."
            )
        if index_directory is None:
            index_directory = pqa_directory("indexes")
        self.index_name = index_name
        self._index_directory = index_directory
        self._schema = None
        self._index = None
        self._searcher = None
        self._index_files: dict[str, str] = {}
        self.changed = False
        self.storage = storage

    async def init_directory(self):
        await anyio.Path(await self.index_directory).mkdir(parents=True, exist_ok=True)

    @staticmethod
    async def extend_and_make_directory(base: anyio.Path, *dirs: str) -> anyio.Path:
        directory = base.joinpath(*dirs)
        await directory.mkdir(parents=True, exist_ok=True)
        return directory

    @property
    async def index_directory(self) -> anyio.Path:
        return await self.extend_and_make_directory(
            anyio.Path(self._index_directory), self.index_name
        )

    @property
    async def index_filename(self) -> anyio.Path:
        return await self.extend_and_make_directory(await self.index_directory, "index")

    @property
    async def docs_index_directory(self) -> anyio.Path:
        return await self.extend_and_make_directory(await self.index_directory, "docs")

    @property
    async def file_index_filename(self) -> anyio.Path:
        return (await self.index_directory) / "files.zip"

    @property
    def schema(self) -> Schema:
        if not self._schema:
            schema_builder = SchemaBuilder()
            for field in self.fields:
                schema_builder.add_text_field(field, stored=True)
            self._schema = schema_builder.build()
        return self._schema

    @property
    async def index(self) -> Index:
        if not self._index:
            index_path = await self.index_filename
            if await (index_path / "meta.json").exists():
                self._index = Index.open(str(index_path))
            else:
                self._index = Index(self.schema, str(index_path))
        return self._index

    @property
    async def searcher(self) -> Searcher:
        if not self._searcher:
            index = await self.index
            index.reload()
            self._searcher = index.searcher()
        return self._searcher

    @property
    async def index_files(self) -> dict[str, str]:
        if not self._index_files:
            file_index_path = await self.file_index_filename
            if await file_index_path.exists():
                async with await anyio.open_file(file_index_path, "rb") as f:
                    content = await f.read()
                    self._index_files = pickle.loads(  # noqa: S301
                        zlib.decompress(content)
                    )
        return self._index_files

    @staticmethod
    def filehash(body: str) -> str:
        return hexdigest(body)

    async def filecheck(self, filename: str, body: str | None = None):
        filehash = None
        if body:
            filehash = self.filehash(body)
        index_files = await self.index_files
        return bool(
            index_files.get(filename)
            and (filehash is None or index_files[filename] == filehash)
        )

    async def add_document(
        self, index_doc: dict, document: Any | None = None, max_retries: int = 1000
    ):
        @retry(
            stop=stop_after_attempt(max_retries),
            wait=wait_random_exponential(multiplier=0.25, max=60),
            retry=retry_if_exception_type(AsyncRetryError),
            reraise=True,
        )
        async def _add_document_with_retry():
            if not await self.filecheck(index_doc["file_location"], index_doc["body"]):
                try:
                    index = await self.index
                    writer = index.writer()
                    writer.add_document(Document.from_dict(index_doc))
                    writer.commit()

                    filehash = self.filehash(index_doc["body"])
                    (await self.index_files)[index_doc["file_location"]] = filehash

                    if document:
                        docs_index_dir = await self.docs_index_directory
                        async with await anyio.open_file(
                            docs_index_dir / f"{filehash}.{self.storage.extension()}",
                            "wb",
                        ) as f:
                            await f.write(self.storage.write_to_string(document))

                    self.changed = True
                except ValueError as e:
                    if "Failed to acquire Lockfile: LockBusy." in str(e):
                        raise AsyncRetryError("Failed to acquire lock") from e
                    raise

        try:
            await _add_document_with_retry()
        except RetryError:
            logger.exception(
                f"Failed to add document after {max_retries} attempts: {index_doc['file_location']}"
            )
            raise

        # Success

    @staticmethod
    @retry(
        stop=stop_after_attempt(1000),
        wait=wait_random_exponential(multiplier=0.25, max=60),
        retry=retry_if_exception_type(AsyncRetryError),
        reraise=True,
    )
    def delete_document(index: Index, file_location: str) -> None:
        try:
            writer = index.writer()
            writer.delete_documents("file_location", file_location)
            writer.commit()
        except ValueError as e:
            if "Failed to acquire Lockfile: LockBusy." in str(e):
                raise AsyncRetryError("Failed to acquire lock") from e
            raise

    async def remove_from_index(self, file_location: str) -> None:
        index_files = await self.index_files
        if index_files.get(file_location):
            index = await self.index
            self.delete_document(index, file_location)
            filehash = index_files.pop(file_location)
            docs_index_dir = await self.docs_index_directory
            # TODO: since the directory is part of the filehash these
            # are always missing. Unsure of how to get around this.
            await (docs_index_dir / f"{filehash}.{self.storage.extension()}").unlink(
                missing_ok=True
            )

            self.changed = True

    async def save_index(self) -> None:
        file_index_path = await self.file_index_filename
        async with await anyio.open_file(file_index_path, "wb") as f:
            await f.write(zlib.compress(pickle.dumps(await self.index_files)))

    async def get_saved_object(
        self, file_location: str, keep_filenames: bool = False
    ) -> Any | None | tuple[Any, str]:
        index_files = await self.index_files
        filehash = index_files.get(file_location)
        if filehash:
            docs_index_dir = await self.docs_index_directory
            async with await anyio.open_file(
                docs_index_dir / f"{filehash}.{self.storage.extension()}", "rb"
            ) as f:
                content = await f.read()
                if keep_filenames:
                    return self.storage.read_from_string(content), file_location
                return self.storage.read_from_string(content)
        return None

    def clean_query(self, query: str) -> str:
        for replace in {"*", "[", "]"}:
            query = query.replace(replace, "")
        return query

    async def query(
        self,
        query: str,
        top_n: int = 10,
        offset: int = 0,
        min_score: float = 0.0,
        keep_filenames: bool = False,
        field_subset: list[str] | None = None,
    ) -> list[Any]:
        query_fields = field_subset or self.fields
        searcher = await self.searcher
        index = await self.index
        results = [
            s[1]
            for s in searcher.search(
                index.parse_query(self.clean_query(query), query_fields), top_n
            ).hits
            if s[0] > min_score
        ][offset : offset + top_n]
        search_index_docs = [searcher.doc(result) for result in results]
        return [
            result
            for result in [
                await self.get_saved_object(
                    doc["file_location"][0], keep_filenames=keep_filenames
                )
                for doc in search_index_docs
            ]
            if result is not None
        ]


async def maybe_get_manifest(filename: anyio.Path | None) -> dict[str, DocDetails]:
    if not filename:
        return {}
    if filename.suffix == ".csv":
        try:
            async with await anyio.open_file(filename, mode="r") as file:
                content = await file.read()
                reader = csv.DictReader(StringIO(content))
                records = [DocDetails(**row) for row in reader]
                return {str(r.file_location): r for r in records if r.file_location}
        except FileNotFoundError:
            logging.warning(f"Manifest file at {filename} could not be found.")
        except Exception:
            logging.exception(f"Error reading manifest file {filename}")
    else:
        logging.error(f"Invalid manifest file type: {filename.suffix}")

    return {}


async def process_file(
    file_path: anyio.Path,
    search_index: SearchIndex,
    metadata: dict[str, Any],
    semaphore: anyio.Semaphore,
    docs_kwargs: dict[str, Any],
) -> None:

    async with semaphore:

        file_name = file_path.name
        if not await search_index.filecheck(str(file_path)):
            logger.info(f"New file to index: {file_name}...")

            doi, title = None, None
            if file_name in metadata:
                doi, title = metadata[file_name].doi, metadata[file_name].title

            # note extras are forbidden in docs
            chunk_chars, overlap = 3000, 250
            if "chunk_chars" in docs_kwargs:
                chunk_chars = int(docs_kwargs.pop("chunk_chars"))
            if "overlap" in docs_kwargs:
                overlap = int(docs_kwargs.pop("overlap"))

            tmp_docs = Docs(**docs_kwargs)
            try:
                await tmp_docs.aadd(
                    path=pathlib.Path(file_path),
                    title=title,
                    doi=doi,
                    chunk_chars=chunk_chars,
                    overlap=overlap,
                    fields=["title", "author", "journal", "year"],
                    use_doc_details=True,
                )
            except ValueError:
                logger.exception(
                    f"Error parsing {file_name}, skipping index for this file."
                )
                (await search_index.index_files)[str(file_path)] = "ERROR"
                await search_index.save_index()
                return

            this_doc = next(iter(tmp_docs.docs.values()))

            if isinstance(this_doc, DocDetails):
                title = this_doc.title or file_name
                year = this_doc.year or "Unknown year"
            else:
                title, year = file_name, "Unknown year"

            await search_index.add_document(
                {
                    "title": title,
                    "year": year,
                    "file_location": str(file_path),
                    "body": "".join([t.text for t in tmp_docs.texts]),
                },
                document=tmp_docs,
            )
            await search_index.save_index()
            logger.info(f"Complete ({title}).")


async def get_directory_index(
    directory: anyio.Path,
    manifest_file: anyio.Path | None = None,
    index_name: str = "pqa_index",
    index_directory: str | os.PathLike | None = None,
    sync_index_w_directory: bool = True,
    use_absolute_directory_path: bool = PQA_INDEX_ABSOLUTE_PATHS,
    max_concurrency: int = 30,
    **docs_kwargs,
) -> SearchIndex:
    """
    Create a Tantivy index from a directory of text files.

    Args:
        directory: Directory to index.
        manifest_file: File with metadata for each document.
        index_name: Name of the index.
        index_directory: Directory to store the index.
        sync_index_w_directory: Sync the index with the directory. (i.e. delete files not in directory)
        use_absolute_directory_path: Use the absolute path for the directory.
        docs_kwargs: Keyword arguments for the Docs object.
        max_concurrency: maximum number of files to be simultaneously indexed.
    """
    semaphore = anyio.Semaphore(max_concurrency)

    if use_absolute_directory_path:
        directory = await directory.absolute()

    search_index = SearchIndex(
        fields=SearchIndex.REQUIRED_FIELDS | {"title", "year"},
        index_name=index_name,
        index_directory=index_directory or pqa_directory("indexes"),
    )

    metadata = await maybe_get_manifest(manifest_file)

    valid_files = [
        file
        async for file in directory.iterdir()
        if file.suffix in {".txt", ".pdf", ".html"}
    ]
    index_files = await search_index.index_files

    if missing := (set(index_files.keys()) - {str(f) for f in valid_files}):
        if sync_index_w_directory:
            for missing_file in missing:
                logger.warning(
                    f"[bold red]Removing {missing_file} from index.[/bold red]"
                )
                await search_index.remove_from_index(missing_file)
            logger.warning("[bold red]Files removed![/bold red]")
        else:
            logger.warning(
                f"[bold red]Indexed files are missing from index folder ({directory}).[/bold red]"
            )
            logger.warning(f"[bold red]files: {missing}[/bold red]")

    async with anyio.create_task_group() as tg:
        for file_path in valid_files:
            if sync_index_w_directory:
                tg.start_soon(
                    process_file,
                    file_path,
                    search_index,
                    metadata,
                    semaphore,
                    docs_kwargs,
                )
            else:
                logger.debug(f"New file {file_path.name} found in directory.")

    if search_index.changed:
        await search_index.save_index()
    else:
        logger.debug("No changes to index.")

    return search_index
